from copy import deepcopy
import logging

import numpy as np

from wepy.reporter.reporter import FileReporter
from wepy.hdf5 import WepyHDF5
from wepy.walker import Walker, WalkerState
from wepy.util.json_top import json_top_atom_count

class WepyHDF5Reporter(FileReporter):
    """Reporter for generating an HDF5 format (WepyHDF5) data file from
    simulations.

    This is the most important reporter as it is the principle output
    format for storing weighted ensemble simulation data.

    Files generated with this reporter can be opened using the
    wepy.hdf5.WepyHDF5 class.

    See Also
    --------

    wepy.hdf5.WepyHDF5


    """

    # this is the name of the dataset that the all atoms will be saved
    # under in the HDF5 alt_reps group
    ALL_ATOMS_REP_KEY = 'all_atoms'

    FILE_ORDER = ('wepy_hdf5_path',)

    # this is the suggested extension for naming WepyHDF5 files used
    # by this reporter, e.g. results.wepy.h5
    SUGGESTED_EXTENSIONS = ('wepy.h5',)

    def __init__(self,
                 save_fields=None,
                 topology=None,
                 units=None,
                 sparse_fields=None,
                 feature_shapes=None, feature_dtypes=None,
                 n_dims=None,
                 main_rep_idxs=None,
                 all_atoms_rep_freq=None,
                 # dictionary of alt_rep keys and a tuple of (idxs, freq)
                 alt_reps=None,

                 # pass in the resampler and boundary
                 # conditions classes to automatically extract the
                 # needed data, the objects themselves are not saves
                 resampler=None,
                 boundary_conditions=None,

                 # or pass the things we need from them in manually
                 resampling_fields=None,
                 decision_enum_dict=None,
                 resampler_fields=None,
                 warping_fields=None,
                 progress_fields=None,
                 bc_fields=None,

                 resampling_records=None,
                 resampler_records=None,
                 warping_records=None,
                 bc_records=None,
                 progress_records=None,

                 # other settings
                 swmr_mode=False,

                 **kwargs
                 ):
        """Constructor for the WepyHDF5Reporter.

        Parameters
        ----------

        save_fields : tuple of str, default: None
           A selection of fields from the walker states to be
           stored. Allows for the ignoring of some states. If None all
           fields from states will attempted to be saved.

        topology : str
            JSON string representing topology of system being simulated.

        units : dict of str: str, optional
            Mapping of trajectory field names to string specs
            for units.

        sparse_fields : list of str, optional
            List of trajectory fields that should be initialized as sparse.

        feature_shapes : dict of str: shape_spec, optional
            Mapping of trajectory fields to their shape spec for initialization.

        feature_dtypes : dict of str: dtype_spec, optional
            Mapping of trajectory fields to their shape spec for initialization.

        n_dims : int, default: 3
            Set the number of spatial dimensions for the default
            positions trajectory field.

        alt_reps : dict of str: tuple of (list of int, int), optional
            Specifies that there will be 'alt_reps' of positions each
            named by the keys of this mapping and containing the
            indices in each value list as the first value of the tuple
            and the second value being the frequency at which this
            field gets saved. Setting `all_atoms_rep_freq` is the
            equivalent of setting an entry {'all_atoms' : ([...],
            `all_atoms_rep_freq`)}.

        main_rep_idxs : list of int, optional
            The indices of atom positions to save as the main 'positions'
            trajectory field. Defaults to all atoms.

        all_atoms_rep_freq : int, optional
            The frequency at which to set an 'alt_rep' for all of the
            atoms in a simulation. Will be set as the field
            'alt_rep/all_atoms'.

        resampler : Resampler object, optional but recommended
            The resampler being used for the simulation. Is used as a
            convenient container for a variety of constants needed for
            specifying data for the resampling records. If this is not
            given then these of the Other Parameters below must be
            specified manually: resampling_fields, decision_enum_dict,
            resampler_fields, resampling_records, resampler_records.

        boundary_conditions : BoundaryConditions object, optional but recommended
            The boundary conditions being used for the simulation. Is
            used as a convenient container for a variety of constants
            needed for specifying data for the warping and progress
            records. If this is not given then these of the Other
            Parameters below must be specified manually:
            warping_fields, progress_fields, bc_fields,
            warping_records, bc_records, progress_records

        swmr_mode : bool
           Whether to write to open the HDF5 in single-writer
           multi-reader (SWMR) mode.


        Other Parameters
        ----------------

        resampling_fields : list of str
            The names of the fields for resampling records

        decision_enum_dict : dict of str : int
            Mapping of the names of resampling decision enum to their
            integer values.

        resampler_fields : list of str
            The names of the fields for the resampler records

        warping_fields : list of str
            The names of the fields for the warping records

        progress_fields : list of str
            The names of the fields for the progress records

        bc_fields : list of str
            The names of the fields for the bounadry condition records.

        resampling_records : list of str, optional
            Names of the resampling_fields that will be used in
            table-like views.

        resampler_records : list of str, optional
            Names of the resampler_fields that will be used in
            table-like views.

        warping_records : list of str, optional
            Names of the warping_fields that will be used in
            table-like views.

        bc_records : list of str, optional
            Names of the bc_fields that will be used in table-like
            views.

        progress_records : list of str, optional
            Names of the progress_fields that will be used in
            table-like views.

        """

        # initialize inherited attributes
        super().__init__(**kwargs)

        # set the preference for swmr mode, True or False, if this is
        # True then SWMR mode will be turned on when the file is
        # written to during reporting
        self.swmr_mode = swmr_mode

        # do all the WepyHDF5 specific stuff

        self.wepy_run_idx = None
        self._tmp_topology = topology
        # which fields from the walker to save, if None then save all of them
        self.save_fields = save_fields
        # dictionary of sparse_field_name -> int : frequency of cycles
        # to save the field
        self._sparse_fields = sparse_fields
        self._feature_shapes = feature_shapes
        self._feature_dtypes = feature_dtypes
        self._n_dims = n_dims

        # get and set the record fields (naems, shapes, dtypes) for
        # the resampler and the boundary conditions
        if (resampling_fields is not None) and (decision_enum_dict is not None):
            self.resampling_fields = resampling_fields
            self.decision_enum = decision_enum_dict
        elif resampler is not None:
            self.resampling_fields = resampler.resampling_fields()
            self.decision_enum = resampler.DECISION.enum_dict_by_name()
        else:
            self.resampling_fields = None
            self.decision_enum = None

        if resampler_fields is not None:
            self.resampler_fields = resampler_fields()
        elif resampler is not None:
            self.resampler_fields = resampler.resampler_fields()
        else:
            self.resampler_fields = None

        if warping_fields is not None:
            self.warping_fields = warping_fields()
        elif boundary_conditions is not None:
            self.warping_fields = boundary_conditions.warping_fields()
        else:
            self.warping_fields = None

        if progress_fields is not None:
            self.progress_fields = progress_fields()
        elif boundary_conditions is not None:
            self.progress_fields = boundary_conditions.progress_fields()
        else:
            self.progress_fields = None

        if bc_fields is not None:
            self.bc_fields = bc_fields()
        elif boundary_conditions is not None:
            self.bc_fields = boundary_conditions.bc_fields()
        else:
            self.bc_fields = None


        # the fields which are records for table like reports
        if resampling_records is not None:
            self.resampling_records = resampling_records
        elif resampler is not None:
            self.resampling_records = resampler.resampling_record_field_names()
        else:
            self.resampling_records = None

        if resampler_records is not None:
            self.resampler_records = resampler_records
        elif resampler is not None:
            self.resampler_records = resampler.resampler_record_field_names()
        else:
            self.resampler_records = None

        if bc_records is not None:
            self.bc_records = bc_records
        elif boundary_conditions is not None:
            self.bc_records = boundary_conditions.bc_record_field_names()
        else:
            self.bc_records = None

        if warping_records is not None:
            self.warping_records = warping_records
        elif boundary_conditions is not None:
            self.warping_records = boundary_conditions.warping_record_field_names()
        else:
            self.warping_records = None

        if progress_records is not None:
            self.progress_records = progress_records
        elif boundary_conditions is not None:
            self.progress_records = boundary_conditions.progress_record_field_names()
        else:
            self.progress_records = None


        # the atom indices of the whole system that will be saved as
        # the main positions representation
        self.main_rep_idxs = main_rep_idxs

        # the idxs for alternate representations of the system
        # positions
        if alt_reps is not None:

            self.alt_reps_idxs = {key: list(tup[0]) for key, tup in alt_reps.items()}

            # add the frequencies for these alt_reps to the
            # sparse_fields frequency dictionary
            for key, (idxs, freq) in alt_reps.items():
                alt_rep_key = "alt_reps/{}".format(key)

                # if the frequency is Ellipsis or 1 then we save it
                # every frame and don't make it sparse because that is
                # very innefficient in comparison
                if freq is Ellipsis or freq == 1 or freq == 0:
                    pass
                else:
                    self._sparse_fields[alt_rep_key] = freq

                self.alt_reps_idxs[key] = list(idxs)

        else:
            self.alt_reps_idxs = {}

        # check for alt_reps of this name because this is reserved for
        # the all_atoms flag.
        if self.ALL_ATOMS_REP_KEY in self.alt_reps_idxs:
            raise ValueError("Cannot name an alt_rep 'all_atoms'")

        # if there is a frequency for all atoms rep then we make an
        # alt_rep for the all_atoms system with the specified
        # frequency
        if all_atoms_rep_freq is not None:
            # count the number of atoms in the topology and set the
            # alt_reps to have the full slice for all atoms
            n_atoms = json_top_atom_count(self._tmp_topology)
            self.alt_reps_idxs[self.ALL_ATOMS_REP_KEY] = np.arange(n_atoms)
            # add the frequency for this sparse fields to the
            # sparse fields dictionary
            self._sparse_fields["alt_reps/{}".format(self.ALL_ATOMS_REP_KEY)] = all_atoms_rep_freq

        # if there are no sparse fields set it as an empty dictionary
        if self._sparse_fields is None:
            self._sparse_fields = {}

        # if units were given add them otherwise set as an empty dictionary
        if units is None:
            self.units = {}
        else:
            self.units = units


    def init(self, continue_run=None,
             init_walkers=None,
             **kwargs):

        # do the inherited stuff
        super().init(**kwargs)

        # open and initialize the HDF5 file
        logging.info("Initializing HDF5 file at {}".format(self.file_path))

        self.wepy_h5 = WepyHDF5(self.file_path, mode=self.mode,
                                topology=self._tmp_topology,
                                units=self.units,
                                sparse_fields=list(self._sparse_fields.keys()),
                                feature_shapes=self._feature_shapes,
                                feature_dtypes=self._feature_dtypes,
                                n_dims=self._n_dims,
                                main_rep_idxs=self.main_rep_idxs,
                                alt_reps=self.alt_reps_idxs)

        # if we specify save fields only save these for the initial walkers
        if self.save_fields is not None:

            state_fields = list(init_walkers[0].state.dict().keys())

            # make sure all the save_fields are present in the state
            assert all([True if save_field in state_fields else False
                        for save_field in self.save_fields]), \
                            "Not all specified save_fields present in walker states"

            filtered_init_walkers = []
            for walker in init_walkers:
                # make a new state by filtering the attributes of the old ones
                state_d = {k : v for k, v in walker.state.dict().items()
                           if k in self.save_fields}

                # and saving alternate representations as we would
                # expect them

                # if there are any alternate representations set them
                for alt_rep_name, alt_rep_idxs in self.alt_reps_idxs.items():

                    alt_rep_path = 'alt_reps/{}'.format(alt_rep_name)

                    # if the idxs are None we want all of the atoms
                    if alt_rep_idxs is None:
                        state_d[alt_rep_path] = state_d['positions'][:]
                    # otherwise get only the atoms we want
                    else:
                        state_d[alt_rep_path] = state_d['positions'][alt_rep_idxs]

                # if the main rep is different then the full state
                # positions set that
                if self.main_rep_idxs is not None:
                    state_d['positions'] = state_d['positions'][self.main_rep_idxs]

                # then making the new state
                new_state = WalkerState(**state_d)

                filtered_init_walkers.append(Walker(new_state, walker.weight))
        # otherwise save the full state
        else:
            filtered_init_walkers = init_walkers

        self.wepy_h5.set_mode(mode='r+')
        with self.wepy_h5:

            # if this is a continuation run of another run we want to
            # initialize it as such

            # initialize a new run
            run_grp = self.wepy_h5.new_run(filtered_init_walkers, continue_run=continue_run)
            self.wepy_run_idx = run_grp.attrs['run_idx']

            # initialize the run record groups using their fields
            self.wepy_h5.init_run_fields_resampling(self.wepy_run_idx, self.resampling_fields)
            # the enumeration for the values of resampling
            self.wepy_h5.init_run_fields_resampling_decision(self.wepy_run_idx, self.decision_enum)
            self.wepy_h5.init_run_fields_resampler(self.wepy_run_idx, self.resampler_fields)
            # set the fields that are records for tables etc. unless
            # they are already set
            if 'resampling' not in self.wepy_h5.record_fields:
                self.wepy_h5.init_record_fields('resampling', self.resampling_records)
            if 'resampler' not in self.wepy_h5.record_fields:
                self.wepy_h5.init_record_fields('resampler', self.resampler_records)

            # if there were no warping fields set there is no boundary
            # conditions and we don't initialize them
            if self.warping_fields is not None:
                self.wepy_h5.init_run_fields_warping(self.wepy_run_idx, self.warping_fields)
                self.wepy_h5.init_run_fields_progress(self.wepy_run_idx, self.progress_fields)
                self.wepy_h5.init_run_fields_bc(self.wepy_run_idx, self.bc_fields)
                # table records
                if 'warping' not in self.wepy_h5.record_fields:
                    self.wepy_h5.init_record_fields('warping', self.warping_records)
                if 'boundary_conditions' not in self.wepy_h5.record_fields:
                    self.wepy_h5.init_record_fields('boundary_conditions', self.bc_records)
                if 'progress' not in self.wepy_h5.record_fields:
                    self.wepy_h5.init_record_fields('progress', self.progress_records)

        # if this was opened in a truncation mode, we don't want to
        # overwrite old runs with future calls to init(). so we
        # change the mode to read/write 'r+'
        if self.mode == 'w':
            self.set_mode(0, 'r+')

    def cleanup(self, **kwargs):

        # it should be already closed at this point but just in case
        if not self.wepy_h5.closed:
            self.wepy_h5.close()

        # remove reference to the WepyHDF5 file so we can serialize this object
        del self.wepy_h5

        super().cleanup(**kwargs)


    def report(self, new_walkers=None,
               cycle_idx=None,
               warp_data=None,
               bc_data=None,
               progress_data=None,
               resampling_data=None,
               resampler_data=None,
               **kwargs):

        n_walkers = len(new_walkers)

        # determine which fields to save. If there were none specified
        # save all of them
        if self.save_fields is None:
            save_fields = list(new_walkers[0].state.dict().keys())
        else:
            save_fields = self.save_fields

        with self.wepy_h5:

            # turn on SWMR mode if requested for reporting
            if self.swmr_mode:
                self.wepy_h5.swmr_mode = True

            # add trajectory data for the walkers
            for walker_idx, walker in enumerate(new_walkers):

                walker_weight = walker.weight
                walker_data = walker.state.dict()

                # iterate through the feature vectors of the walker
                # (fields), and the keys for the alt_reps
                for field_path in list(walker_data.keys()):

                    # save the field if it is in the list of save_fields
                    if field_path not in save_fields:
                        walker_data.pop(field_path)
                        continue

                    # if the result is None don't save anything
                    if walker_data[field_path] is None:
                        walker_data.pop(field_path)
                        continue

                    # if this is a sparse field we decide
                    # whether it is a valid cycle to save on
                    if field_path in self._sparse_fields:
                        if cycle_idx % self._sparse_fields[field_path] != 0:
                            # this is not a valid cycle so we
                            # remove from the walker_data
                            walker_data.pop(field_path)
                            continue


                # Add the alt_reps fields by slicing the positions
                for alt_rep_key, alt_rep_idxs in self.alt_reps_idxs.items():
                    alt_rep_path = "alt_reps/{}".format(alt_rep_key)

                    # if the alt rep is also a sparse field check this
                    if alt_rep_path in self._sparse_fields:

                        # check to make sure this is a cycle this is
                        # to be saved to, if it is not continue on to
                        # the next field without saving this one
                        if cycle_idx % self._sparse_fields[alt_rep_path] != 0:

                            continue

                    # slice them and save them

                    # if the idxs are None we want all of the atoms
                    if alt_rep_idxs is None:
                        alt_rep_data = walker_data['positions'][:]
                    # otherwise get only th atoms we want
                    else:
                        alt_rep_data = walker_data['positions'][alt_rep_idxs]
                    walker_data[alt_rep_path] = alt_rep_data


                # lastly reduce the atoms for the main representation
                # if this option was given
                if self.main_rep_idxs is not None:
                    walker_data['positions'] = walker_data['positions'][self.main_rep_idxs]


                # for all of these fields we wrap them in another
                # dimension to make them feature vectors
                for field_path in list(walker_data.keys()):
                    walker_data[field_path] = np.array([walker_data[field_path]])

                # save the data to the HDF5 file for this walker

                # check to see if the walker has a trajectory in the run
                if walker_idx in self.wepy_h5.run_traj_idxs(self.wepy_run_idx):

                    # if it does then append to the trajectory
                    self.wepy_h5.extend_traj(self.wepy_run_idx, walker_idx,
                                             weights=np.array([[walker_weight]]),
                                             data=walker_data)
                # start a new trajectory
                else:
                    # add the traj for the walker with the data

                    traj_grp = self.wepy_h5.add_traj(self.wepy_run_idx,
                                                     weights=np.array([[walker_weight]]),
                                                     data=walker_data)

                    # add as metadata the cycle idx where this walker started
                    traj_grp.attrs['cycle_idx'] = cycle_idx


            # report the boundary conditions records data, if boundary
            # conditions were initialized
            if self.warping_fields is not None:
                self._report_warping(cycle_idx, warp_data)
                self._report_bc(cycle_idx, bc_data)
                self._report_progress(cycle_idx, progress_data)

            # report the resampling records data
            self._report_resampling(cycle_idx, resampling_data)

            self._report_resampler(cycle_idx, resampler_data)

        super().report(**kwargs)


    # sporadic
    def _report_warping(self, cycle_idx, warping_data):
        """Method to write warping specific information.

        Parameters
        ----------
        cycle_idx : int

        warp_data : list of dict of str : value
            List of dict-like records for each warping event from the
            last cycle.

        """

        if len(warping_data) > 0:
            self.wepy_h5.extend_cycle_warping_records(self.wepy_run_idx, cycle_idx, warping_data)

    def _report_bc(self, cycle_idx, bc_data):
        """Method to write boundary condition update specific information.

        Parameters
        ----------
        cycle_idx : int

        bc_data : list of dict of str : value
           List of dict-like records specifying the changes to the
           state of the boundary conditions in the last cycle.

        """

        if len(bc_data) > 0:
            self.wepy_h5.extend_cycle_bc_records(self.wepy_run_idx, cycle_idx, bc_data)

    def _report_resampler(self, cycle_idx, resampler_data):
        """Method to write resampler update specific information.

        Parameters
        ----------
        cycle_idx : int

        resampler_data : list of dict of str : value
            List of records specifying the changes to the state of the
            resampler in the last cycle.

        """

        if len(resampler_data) > 0:
            self.wepy_h5.extend_cycle_resampler_records(self.wepy_run_idx, cycle_idx, resampler_data)

    # the resampling records are provided every cycle but they need to
    # be saved as sporadic because of the variable number of walkers
    def _report_resampling(self, cycle_idx, resampling_data):
        """Method to write resampling specific information.

        Parameters
        ----------
        cycle_idx : int

        resampling_data : list of dict of str : value
            List of records specifying the resampling to occur at this
            cycle.

        """

        self.wepy_h5.extend_cycle_resampling_records(self.wepy_run_idx, cycle_idx, resampling_data)

    # continual
    def _report_progress(self, cycle_idx, progress_data):
        """Method to write progress specific information.

        Parameters
        ----------
        cycle_idx : int

        progress_data : dict str : list
            A record indicating the progress values for each walker in
            the last cycle.

        """

        self.wepy_h5.extend_cycle_progress_records(self.wepy_run_idx, cycle_idx, [progress_data])






"""Implement a key-value store on top of sqlite3 database."""

# Copyright (c) 2012, Alex Morega
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Software copied and modified heavily from this source


import os
import os.path as osp
import logging
import sqlite3
from collections.abc import MutableMapping
from contextlib import contextmanager

# mapping of the modes we support and the modes that SQLite provides
# KV mode -> sqlite3 mode
MODE_MAPPING = (
    ('r', 'ro'),
    ('r+', 'rw'),
    ('a', 'rwc'),
    ('x', None),
    ('w', None),
    ('w-', None),
)

# modes that sqlite3 itself doesn't support and have to manual
# processing for
TRUNCATE_MODES = ('w',)
FAIL_IF_EXISTS_CREATE_MODES = ('w-', 'x')

SQLITE3_URI_TEMPLATE = "{protocol}:{url}"
SQLITE3_QUERY_URI_TEMPLATE = "{protocol}:{url}?{query}"

SQLITE3_QUERY_JOIN_CHAR = "&"

SQLITE3_INMEMORY_URI = "file::memory:?cache=shared"


# the default types for the values that is checked, is bytes
DEFAULT_VALUE_TYPES = (bytes, bytearray)

def gen_uri(db_url, mode_spec):

    # if the db url is the in memory special string or None or the
    # :memory: identifier, use the full in-memory URI
    if (db_url == SQLITE3_INMEMORY_URI or
        db_url is None or
        db_url == ":memory:"):

        db_uri = SQLITE3_INMEMORY_URI

        return db_uri

    # check for and split the URI into the components: protocol, URL, query
    if len(db_url.split(':')) > 1:
        # for a query
        if len(db_url.split('?')) > 1:

            # protocol and queries
            ((protocol,), (url, query)) = [comp.split('?') for comp in
                                           db_url.split(':')]

        else:
            # no query
            protocol, url = db_url.split(':')
            query = ""
    else:
        protocol = ""
        if len(db_url.split('?')) > 1:
            url, query = db_url.split('?')
        else:
            query = ""
            url = db_url

    # process and build the URI given this information

    # if the protocol is not given set it to the default
    if len(protocol) == 0:
        protocol = 'file'

    # split the query up into sections if it has them
    if len(query.split('?')) > 1:
        queries = query.split('?')

        # split the "key=value" pairs in each query
        queries = {q.split('=')[0] : q.split('=')[0]
                   for q in queries}
    else:
        queries = {}

    # now handle the mode. If it was given as an argument then we have
    # to set the appropriate query in the URI. If it was given in the
    # URI that takes precedence however.

    # check for a mode option in the given query section, if no mode
    # is given then we set it depending on what mode_spec was given
    if 'mode' not in queries:

        # default to rwc mode
        mode_query_value = 'rwc'

        if mode_spec is not None:

            # if the mode is one of the modes in the mode mapping use that to
            # generate the URI, otherwise raise an error
            if mode_spec not in dict(MODE_MAPPING):

                raise ValueError("kv mode spec '{}' not recognized".format(mode_spec))

            else:

                sqlite_mode = dict(MODE_MAPPING)[mode_spec]

                # if the sqlite_mode is recognized as a mode that can
                # be given in the query section do that
                if sqlite_mode is not None:
                    mode_query_value = sqlite_mode

                # otherwise we handle these special modes ourselves
                # here, checking file properties and raising errors as
                # necessary
                else:

                    # if the mode is either 'x' or 'w-' check to see if the db
                    # already exists. If it does raise an error.
                    if mode_spec in FAIL_IF_EXISTS_CREATE_MODES:

                        if osp.exists(url):
                            raise OSError("File exists")

                    # if it is 'w' we want to delete the old file
                    elif mode_spec in TRUNCATE_MODES:

                        # if it exists remove it
                        if osp.exists(url):
                            os.remove(url)

                    # the sqlite_mode for these is rwc (a), since we
                    # need to create it
                    mode_query_value = "rwc"

        # add the mode query to queries list
        queries['mode'] = mode_query_value

    # if thw queries are empty just use the protocol and URL
    if not queries:
        db_uri = SQLITE3_URI_TEMPLATE.format(protocol=protocol,
                                             url=url)
    # otherwise do the whole thing
    else:

        # build the query substring
        query = SQLITE3_QUERY_JOIN_CHAR.join(["{}={}".format(key, value)
                                              for key, value in queries.items()])

        # build the URI string
        db_uri = SQLITE3_QUERY_URI_TEMPLATE.format(protocol=protocol,
                                                   url=url,
                                                   query=query)

    return db_uri

class KV(MutableMapping):

    def __init__(self, db_url=None,
                 table='data',
                 primary_key='key',
                 value_name='value',
                 timeout=5,
                 mode='x',
                 append_only=False,
                 value_types=DEFAULT_VALUE_TYPES,
    ):

        # generate a good URI from the url and the mode
        db_uri = gen_uri(db_url, mode)

        self._mode = mode
        self._append_only = append_only

        # set the value types for this kv
        self._kv_types = value_types


        self._db_uri = db_uri

        # connect to the db
        self._db = sqlite3.connect(self._db_uri, timeout=timeout, uri=True)
        self._closed = False

        # set the isolation level to autocommit
        self._db.isolation_level = None

        # we can use read_uncommited only in append_only mode (no
        # updates) because you never have to worry about dirty reads
        # since you can't update
        if self.append_only:
            self._execute("PRAGMA read_uncommited=1")


        self._table = table
        self._primary_key = primary_key
        self._value_name = value_name

        # create the table if it doesn't exist and set the key names
        create_table_query = """
            CREATE TABLE IF NOT EXISTS {table_name}
            ({key_name} PRIMARY KEY, {value_name})
        """.format(table_name=self.table,
                   key_name=self.primary_key,
                   value_name=self.value_name)
        self._execute(create_table_query)

        self._locks = 0

    @property
    def mode(self):
        return self._mode

    @property
    def append_only(self):
        return self._append_only

    def close(self):

        if self._closed == True:
            raise IOError("The database connection is already closed")

        else:
            self._db.close()
            self._closed = True


    @property
    def db_uri(self):
        return self._db_uri

    @property
    def db(self):
        return self._db

    @property
    def table(self):
        return self._table

    @property
    def primary_key(self):
        return self._primary_key

    @property
    def value_name(self):
        return self._value_name

    @property
    def value_types(self):
        return self._kv_types


    def _execute(self, *args):
        return self._db.cursor().execute(*args)

    def __len__(self):
        [[n]] = self._execute('SELECT COUNT(*) FROM {table}'.format(table=self.table))
        return n

    def __getitem__(self, key):

        if key is None:
            query = ('SELECT {value} FROM {table} WHERE {key} is NULL'.format(
                     value=self.value_name,
                     table=self.table,
                     key=self.primary_key), ())
        else:
            query = ('SELECT {value} FROM {table} WHERE {key}=?'.format(
                                                      value=self.value_name,
                                                      table=self.table,
                                                      key=self.primary_key),
                     (key,),)

        cursor = self._execute(*query)
        result = cursor.fetchone()

        if result is None:
            raise KeyError
        else:
            return result[0]

    def __iter__(self):
        return (key for [key] in self._execute('SELECT {key} FROM {table}'.format(
                                               key=self.primary_key,
                                               table=self.table),
                                               ()
                                              )
                )

    def __setitem__(self, key, value):
        """Set a value, must be in bytes format."""

        # check the type of the value to make sure it is what this KV
        # supports, if it is None then it is the standard python type
        # translation

        if self.value_types is not None:
            assert isinstance(value, self.value_types), "Value must be a value supported by this kv"

        self.lockless_set(key, value)


    def __delitem__(self, key):

        logging.debug("Deleting the snapshot {}".format(key))

        # no deletions in append only mode
        if self.append_only:
            raise sqlite3.IntegrityError("DB is opened in append only mode, "
                                         "and {} has already been set".format(key))

        # delete it if it exists
        elif key in self:
            logging.debug("executing delete query")
            self._execute(self.del_query,
                          (key,))
            logging.debug("finished")

        else:
            raise KeyError

    @property
    def insert_query(self):
        query = 'INSERT INTO {table} VALUES (?, ?)'.format(table=self.table)

        return query

    @property
    def update_query(self):

        query = 'UPDATE {table} SET {value}=? WHERE {key}=?'.format(
            key=self.primary_key,
            value=self.value_name,
            table=self.table)

        return query

    @property
    def del_query(self):

        query = 'DELETE FROM {table} WHERE {key}=?'.format(
            key=self.primary_key,
            table=self.table)

        return query

    def lockless_set(self, key, value):
        """an implementation of the __setitem__ without the lock context
        manager which turns on the DEFERRED isolation level. The
        isolation level of the KV is set to autocommit so now lock is
        needed anyhow.

        """

        # insert the key-value pair if the key isn't in the db
        try:
            self._execute(self.insert_query, (key, value))

        # otherwise update the keys value
        except sqlite3.IntegrityError:

            # if we are in append only mode don't allow updates
            if self.append_only:
                raise sqlite3.IntegrityError("DB is opened in append only mode, "
                                             "and {} has already been set".format(key))
            else:
                self._execute(self.update_query,
                              (value, key))

    def set_in_tx(self, cursor, key, value):
        """Do a set with a cursor, this allows it to be done in a transaction."""

        try:
            cursor.execute(self.insert_query, (key, value))
        except sqlite3.IntegrityError:

            # if we are in append only mode don't allow updates
            if self.append_only:
                raise sqlite3.IntegrityError("DB is opened in append only mode, "
                                             "and {} has already been set".format(key))

            else:
                cursor.execute(self.update_query, (key, value))

        return cursor

    def del_in_tx(self, cursor, key):

        # no deletions in append only mode
        if self.append_only:
            raise sqlite3.IntegrityError("DB is opened in append only mode, "
                                         "and {} has already been set".format(key))

        elif key in self:
            cursor.execute(self.del_query, (key,))

        else:
            raise KeyError

        return cursor


    @contextmanager
    def lock(self):
        if not self._locks:
            self._execute('BEGIN TRANSACTION')
        self._locks = True
        try:
            yield
        finally:
            self._locks = False
            if not self._locks:
                self._execute('COMMIT')



import multiprocessing as mulproc
import random as rand
import itertools as it

import logging
from eliot import start_action, log_call

import numpy as np

from wepy.resampling.resamplers.resampler import Resampler
from wepy.resampling.resamplers.clone_merge  import CloneMergeResampler
from wepy.resampling.decisions.clone_merge import MultiCloneMergeDecision

class REVOResampler(CloneMergeResampler):
    r"""Resampler implementing the REVO algorithm.

    You can find more detailed information in the paper "REVO:
    Resampling of ensembles by variation optimization" but
    briefly:

    REVO is a Weighted Ensemble based enhanced sampling algorithm
    which uses cloning and merging to create ensembles of diverse
    trajectories without defining any regions. It instead optimizes a
    measure of “variation” that depends on the pairwise distances
    between the walkers and their weights.


    REVO solves this optimization problem using a greedy algorithm
    which at each step selects best walkers for resampling operations
    (cloning and merging) in order to maximize the "trajectory
    variation".

    The trajectory variation is defined as


    .. math::
        V = \sum_{i} V_i = \sum_i \sum_{j}(\frac{d_{ij}}{d_0})
        ^{\alpha}\phi_i\phi_j

    where

    :math:`V_i` : the trajectory variation value of walker `i`

    :math:`d_{ij}` : the distance between walker i and j
    according the distance metric

    :math:`\alpha` : modulates the influence of the distances in the
    variation calculation

    :math:`d_0` : the characteristic distance and is used to make the
    equation unitless.

    :math:`\phi` : is a non-negative function which is a measure of
    the relative importance of the walker and is referred to as a
    "novelty function".  Here it is a function of a walker's weight.

    Furthermore REVO needs the following parameters:

       pmin: the minimum statistical weight. REVO does not clone
       walkers with a weight less than pmin.

       pmax: The maximum statistical weight. It prevents the
       accumulation of too much weight in one walker.

       merge_dist: This is the merge-distance threshold. The distance
        between merged walkers should be less than this value.

    The resample function, called during every cycle, takes the
    ensemble of walkers and performs the follow steps:

       - Calculate the pairwise all-to-all distance matrix using the distance metric
       - Decides which walkers should be merged or cloned
       - Applies the cloning and merging decisions to get the resampled walkers
       - Creates the resampling data that includes
       - distance_matrix : the calculated all-to-all distance matrix

           - n_walkers : the number of walkers. number of walkers is
             kept constant thought the resampling.

           - variation : the final value of trajectory variation

           - images : the images of walkers that is defined by the distance object

           - image_shape : the shape of the image

    The algorithm saves the records of cloning and merging
    information in resampling data.

    Only the net clones and merges are recorded in the resampling records.

    """

    # fields for resampler data
    RESAMPLING_FIELDS = CloneMergeResampler.RESAMPLING_FIELDS
    RESAMPLING_SHAPES = CloneMergeResampler.RESAMPLING_SHAPES #+ (Ellipsis,)
    RESAMPLING_DTYPES = CloneMergeResampler.RESAMPLING_DTYPES #+ (np.int,)


    # fields that can be used for a table like representation
    RESAMPLING_RECORD_FIELDS = CloneMergeResampler.RESAMPLING_RECORD_FIELDS

    # fields for resampling data
    RESAMPLER_FIELDS = CloneMergeResampler.RESAMPLER_FIELDS + \
                       ('num_walkers', 'distance_matrix', 'variation', 'image_shape', 'images')
    RESAMPLER_SHAPES = CloneMergeResampler.RESAMPLER_SHAPES + \
                       ((1,), Ellipsis, (1,), Ellipsis, Ellipsis)
    RESAMPLER_DTYPES = CloneMergeResampler.RESAMPLER_DTYPES + \
                       (np.int, np.float, np.float, np.int, None)

    # fields that can be used for a table like representation
    RESAMPLER_RECORD_FIELDS = CloneMergeResampler.RESAMPLER_RECORD_FIELDS + \
                              ('variation',)


    def __init__(self,
                 merge_dist=None,
                 char_dist=None,
                 distance=None,
                 init_state=None,
                 weights=True,
                 pmin=1e-12,
                 pmax=0.1,
                 dist_exponent=4,
                 seed=None,
                 **kwargs):

        """Constructor for the REVO Resampler.

        Parameters
        ----------

        dist_exponent : int
          The distance exponent that modifies distance and weight novelty
          relative to each other in the variation equation.

        merge_dist : float
            The merge distance threshold. Units should be the same as
            the distance metric.

        char_dist : float
            The characteristic distance value. It is calculated by
            running a single dynamic cycle and then calculating the
            average distance between all walkers. Units should be the
            same as the distance metric.

        distance : object implementing Distance
            The distance metric to compare walkers.

        weights : bool
            Turns off or on the weight novelty in
            calculating the variation equation. When weight is
            False, the value of the novelty function is set to 1 for all
            walkers.

        init_state : WalkerState object
            Used for automatically determining the state image shape.

        seed : None or int, optional
            The random seed. If None, the system (random) one will be used.

        """

        # call the init methods in the CloneMergeResampler
        # superclass. We set the min and max number of walkers to be
        # constant
        super().__init__(pmin=pmin, pmax=pmax,
                         min_num_walkers=Ellipsis,
                         max_num_walkers=Ellipsis,
                         **kwargs)

        assert merge_dist is not None, "Merge distance must be given."
        assert distance is not None,  "Distance object must be given."
        assert char_dist is not None, "Characteristic distance value (d0) must be given"
        assert init_state is not None,  "An initial state must be given."

        # ln(probability_min)
        self.lpmin = np.log(self.pmin/100)
        self.dist_exponent = dist_exponent

        # the distance metric

        self.merge_dist = merge_dist

        # the distance metric

        self.distance = distance

        # the characteristic distance, char_dist

        self.char_dist = char_dist

        # setting the random seed
        self.seed = seed
        if seed is not None:
            rand.seed(seed)

        # setting the weights parameter
        self.weights = weights

        # we do not know the shape and dtype of the images until
        # runtime so we determine them here

        image = self.distance.image(init_state)
        self.image_dtype = image.dtype

    def resampler_field_dtypes(self):
        """ Finds out the datatype of the image.

        Returns
        -------
        datatypes : tuple of datatype
        The type of reasampler image.

        """

        # index of the image idx
        image_idx = self.resampler_field_names().index('images')

        # dtypes adding the image dtype
        dtypes = list(super().resampler_field_dtypes())
        dtypes[image_idx] = self.image_dtype

        return tuple(dtypes)

    def _novelty(self, walker_weight, num_walker_copy):
        """Calculates the novelty fuction value.

        Parameters
        ----------

        walker_weight : float
            The weight of the walker.

        num_walker_copy : int
          The number of copies of the walker.

        Returns
        -------
        novelty : float
        The calcualted value of novelty for the given walker.

        """

        novelty = 0

        if walker_weight > 0 and num_walker_copy > 0:

            if self.weights:

                novelty = np.log(walker_weight/num_walker_copy) - self.lpmin

            else:

                novelty = 1

        if novelty < 0:

            novelty = 0

        return novelty

    def _calcvariation(self, walker_weights, num_walker_copies, distance_matrix):
        """Calculates the variation value.

        Parameters
        ----------

        walker_weights : list of float
            The weights of all walkers. The sum of all weights should be 1.0.

        num_walker_copies : list of int
            The number of copies of each walker.
            0 means the walker is not exists anymore.
            1 means there is one of the this walker.
            >1 means it should be cloned to this number of walkers.

        distance_matrix : list of arraylike of shape (num_walkers)

        Returns
        -------
        variation : float
           The calculated variation value.

        walker_variations : arraylike of shape (num_walkers)
           The Vi value of each walker.

        """

        num_walkers = len(walker_weights)



        # set the novelty values
        walker_novelties = np.array([self._novelty(walker_weights[i], num_walker_copies[i])
                               for i in range(num_walkers)])


        # the value to be optimized
        variation = 0

        # the walker variation values (Vi values)
        walker_variations = np.zeros(num_walkers)


        # calculate the variation and walker variation values
        for i in range(num_walkers - 1):

            if num_walker_copies[i] > 0:
                for j in range(i+1, num_walkers):

                    if num_walker_copies[j] > 0:

                        partial_variation = ((distance_matrix[i][j] / self.char_dist) ** self.dist_exponent) \
                        * walker_novelties[i] * walker_novelties[j]

                        variation += partial_variation * num_walker_copies[i] * num_walker_copies[j]
                        walker_variations[i] += partial_variation * num_walker_copies[j]
                        walker_variations[j] += partial_variation * num_walker_copies[i]

        return variation, walker_variations

    def decide(self, walker_weights, num_walker_copies, distance_matrix):
        """Optimize the trajectory variation by making decisions for resampling.

        Parameters
        ----------

        walker_weights : list of flaot
            The weights of all walkers. The sum of all weights should be 1.0.

        num_walker_copies : list of int
            The number of copies of each walker.
            0 means the walker is not exists anymore.
            1 means there is one of the this walker.
            >1 means it should be cloned to this number of walkers.

        distance_matrix : list of arraylike of shape (num_walkers)

        Returns
        -------
        variation : float
            The optimized value of the trajectory variation.

        resampling_data : list of dict of str: value
            The resampling records resulting from the decisions.

        """
        num_walkers = len(walker_weights)

        variations = []
        merge_groups = [[] for i in range(num_walkers)]
        walker_clone_nums = [0 for i in range(num_walkers)]

        # make copy of walkers properties
        new_walker_weights = walker_weights.copy()
        new_num_walker_copies = num_walker_copies.copy()


        # calculate the initial variation which will be optimized
        variation, walker_variations = self._calcvariation(walker_weights,
                                                           new_num_walker_copies,
                                                           distance_matrix)
        variations.append(variation)

        # maximize the variance through cloning and merging
        logging.info("Starting variance optimization: {}".format(variation))

        productive = True
        while productive:
            productive = False
            # find min and max walker_variationss, alter new_amp

            # initialize to None, we may not find one of each
            min_idx = None
            max_idx = None

            # selects a walker with minimum walker_variations and a walker with
            # maximum walker_variations walker (distance to other walkers) will be
            # tagged for cloning (stored in maxwind), except if it is
            # already a keep merge target
            max_tups = []
            for i, value in enumerate(walker_variations):
                # 1. must have an amp >=1 which gives the number of clones to be made of it
                # 2. clones for the given amplitude must not be smaller than the minimum probability
                # 3. must not already be a keep merge target
                if (new_num_walker_copies[i] >= 1) and \
                   (new_walker_weights[i]/(new_num_walker_copies[i] + 1) > self.pmin) and \
                   (len(merge_groups[i]) == 0):
                    max_tups.append((value, i))


            if len(max_tups) > 0:
                max_value, max_idx = max(max_tups)

            # walker with the lowest walker_variations (distance to other walkers)
            # will be tagged for merging (stored in min_idx)
            min_tups = [(value, i) for i,value in enumerate(walker_variations)
                        if new_num_walker_copies[i] == 1 and (new_walker_weights[i]  < self.pmax)]

            if len(min_tups) > 0:
                min_value, min_idx = min(min_tups)

            # does min_idx have an eligible merging partner?
            # closedist = self.merge_dist
            closewalk = None
            condition_list = np.array([i is not None for i in [min_idx, max_idx]])
            if condition_list.all() and min_idx != max_idx:

                # get the walkers that aren't the minimum and the max
                # walker_variations walkers, as candidates for merging
                closewalks = set(range(num_walkers)).difference([min_idx, max_idx])

                # remove those walkers that if they were merged with
                # the min walker_variations walker would violate the pmax
                closewalks = [idx for idx in closewalks
                                      if (new_num_walker_copies[idx]==1) and
                                       (new_walker_weights[idx] + new_walker_weights[min_idx] < self.pmax)
                                      ]

                # if there are any walkers left, get the distances of
                # the close walkers to the min walker_variations walker if that
                # distance is less than the maximum merge distance
                if len(closewalks) > 0:
                    closewalks_dists = [(distance_matrix[min_idx][i], i) for i in closewalks
                                            if distance_matrix[min_idx][i] < (self.merge_dist)]

                    # if any were found set this as the closewalk
                    if len(closewalks_dists) > 0:
                        closedist, closewalk = min(closewalks_dists)


            # did we find a closewalk?
            condition_list = np.array([i is not None for i in [min_idx, max_idx, closewalk]])
            #if we find a walker for cloning, a walker and its close neighbor for merging
            if condition_list.all() :

                # change new_amp
                tempsum = new_walker_weights[min_idx] + new_walker_weights[closewalk]
                new_num_walker_copies[min_idx] = new_walker_weights[min_idx]/tempsum
                new_num_walker_copies[closewalk] = new_walker_weights[closewalk]/tempsum
                new_num_walker_copies[max_idx] += 1

                # re-determine variation function, and walker_variations values
                new_variation, walker_variations = self._calcvariation(new_walker_weights, new_num_walker_copies, distance_matrix)

                if new_variation > variation:
                    variations.append(new_variation)

                    logging.info("Variance move to {} accepted".format(new_variation))

                    productive = True
                    variation = new_variation

                    # make a decision on which walker to keep
                    # (min_idx, or closewalk), equivalent to:
                    # `random.choices([closewalk, min_idx],
                    #                 weights=[new_walker_weights[closewalk], new_walker_weights[min_idx])`
                    r = rand.uniform(0.0, new_walker_weights[closewalk] + new_walker_weights[min_idx])

                     # keeps closewalk and gets rid of min_idx
                    if r < new_walker_weights[closewalk]:
                        keep_idx = closewalk
                        squash_idx = min_idx

                    # keep min_idx, get rid of closewalk
                    else:
                        keep_idx = min_idx
                        squash_idx = closewalk

                    # update weight
                    new_walker_weights[keep_idx] += new_walker_weights[squash_idx]
                    new_walker_weights[squash_idx] = 0.0

                    # update new_num_walker_copies
                    new_num_walker_copies[squash_idx] = 0
                    new_num_walker_copies[keep_idx] = 1

                    # add the squash index to the merge group
                    merge_groups[keep_idx].append(squash_idx)

                    # add the indices of the walkers that were already
                    # in the merge group that was just squashed
                    merge_groups[keep_idx].extend(merge_groups[squash_idx])

                    # reset the merge group that was just squashed to empty
                    merge_groups[squash_idx] = []

                    # increase the number of clones that the cloned
                    # walker has
                    walker_clone_nums[max_idx] += 1

                    # new variation for starting new stage
                    new_variation, walker_variations = self._calcvariation(new_walker_weights,
                                                                          new_num_walker_copies,
                                                                          distance_matrix)
                    variations.append(new_variation)

                    logging.info("variance after selection: {}".format(new_variation))

                # if not productive
                else:
                    new_num_walker_copies[min_idx] = 1
                    new_num_walker_copies[closewalk] = 1
                    new_num_walker_copies[max_idx] -= 1

        # given we know what we want to clone to specific slots
        # (squashing other walkers) we need to determine where these
        # squashed walkers will be merged
        walker_actions = self.assign_clones(merge_groups, walker_clone_nums)

        # because there is only one step in resampling here we just
        # add another field for the step as 0 and add the walker index
        # to its record as well
        for walker_idx, walker_record in enumerate(walker_actions):
            walker_record['step_idx'] = np.array([0])
            walker_record['walker_idx'] = np.array([walker_idx])

        return walker_actions, variations[-1]

    def _all_to_all_distance(self, walkers):
        """ Calculate the pairwise all-to-all distances between walkers.

        Parameters
        ----------
        walkers : list of walkers


        Returns
        -------
        distance_matrix : list of arraylike of shape (num_walkers)

        images : list of image obeject

        """
        # initialize an all-to-all matrix, with 0.0 for self distances
        dist_mat = np.zeros((len(walkers), len(walkers)))

        # make images for all the walker states for us to compute distances on
        images = []
        for walker in walkers:
            image = self.distance.image(walker.state)
            images.append(image)

        # get the combinations of indices for all walker pairs
        for i, j in it.combinations(range(len(images)), 2):

            # calculate the distance between the two walkers
            dist = self.distance.image_distance(images[i], images[j])

            # save this in the matrix in both spots
            dist_mat[i][j] = dist
            dist_mat[j][i] = dist

        return [walker_dists for walker_dists in dist_mat], images

    @log_call(include_args=[],
              include_result=False)
    def resample(self, walkers):
        """Resamples walkers based on REVO algorithm

        Parameters
        ----------
        walkers : list of walkers


        Returns
        -------
        resampled_walkers : list of resampled_walkers

        resampling_data : list of dict of str: value
            The resampling records resulting from the decisions.

        resampler_data :list of dict of str: value
            The resampler records resulting from the resampler actions.

        """

        #initialize the parameters
        num_walkers = len(walkers)
        walker_weights = [walker.weight for walker in walkers]
        num_walker_copies = [1 for i in range(num_walkers)]

        # calculate distance matrix
        distance_matrix, images = self._all_to_all_distance(walkers)

        logging.info("distance_matrix")
        logging.info("\n{}".format(str(np.array(distance_matrix))))

        # determine cloning and merging actions to be performed, by
        # maximizing the variation, i.e. the Decider
        resampling_data, variation = self.decide(walker_weights, num_walker_copies, distance_matrix)

        # convert the target idxs and decision_id to feature vector arrays
        for record in resampling_data:
            record['target_idxs'] = np.array(record['target_idxs'])
            record['decision_id'] = np.array([record['decision_id']])

        # actually do the cloning and merging of the walkers
        resampled_walkers = self.DECISION.action(walkers, [resampling_data])

       # flatten the distance matrix and give the number of walkers
        # as well for the resampler data, there is just one per cycle
        resampler_data = [{'distance_matrix' : np.ravel(np.array(distance_matrix)),
                           'num_walkers' : np.array([len(walkers)]),
                           'variation' : np.array([variation]),
                           'images' : np.ravel(np.array(images)),
                           'image_shape' : np.array(images[0].shape)}]

        return resampled_walkers, resampling_data, resampler_data



"""Classes used for efficiently performing work at different stages of
the wepy simulation cycle.

Wepy simulations have four sequential stages of computation:

1. running segments
2. boundary conditions
3. resampling
4. reporting

Stage 1 involves propagating dynamics for each walker in a completely
independent using the same runner function (i.e. `run_segment`) and
can always be completely parallelized.

Steps 2 and 4, with a few more assumptions of independence (which are
most likely satisfied without overly complex behavior), are probably
also completely parallelizable over walkers and reporters
respectively.

Resampling may also have a degree of parallelizability but will vary
considerably between resamplers and so cannot be treated outside of
any individual framework for resamplers and so will not be treated
generally in this submodule.

Currently, wepy supports a worker based task queue implementation
(wepy.work_mapper.mapper.WorkerMapper) for running parallel segments
in stage 1, which is very useful for utilizing GPU compute sources.
It is pretty general and utilizes the multiprocessing library for
starting worker processes. Specific code for running OpenMM GPU
calculations is contained within the openmm runner module.

Currently general solutions for parallelizing stage 2 and 3 are not
provided or supported in the simulation manager (to avoid premature
optimization) but these are easy targets for improving the efficiency
of wepy code.

If you don't care about parallelizing your segment running (say for
simple test systems) you can use the reference implementation
(wepy.work_mapper.mapper.Mapper) which is basically just a wrapper
around a for-loop.

This sub-module provides reference implementations and/or abstract
base classes for a few interfaces.

The WorkMapper class interface has the following methods:

- init
- cleanup
- map

A single attribute for the segment_function that is set at runtime
with a call to `init`:

- segment_func

And an optional attribute to enable reporting on segment running performance:

- worker_segment_times : dict of int : list of float

Which should be a dictionary mapping a worker index to a list of float
values of time (in seconds) for each task (i.e. segment of walker
dynamics) it ran in the last cycle only (not cumulative over
consecutive cycles).

The `init` method is called at runtime by the simulation manager at
the beginning of the simulation and allows for performing such actions
as opening file handles or starting worker processes (as is the case
with the WorkerMapper).

The only necessary key-word argument to this method is 'segment_func'
which will be provided by the simulation manager (and is derived at
runtime from the runner).

The `cleanup` method likewise is called either at the end of a
succesful simulation or when an error occurs in the call to
`run_cycle` in the simulation manager (e.g. to allow killing of live
processes).

The `map` function acts similar to the python builtin except it does
not accept the function to map over the data. This is instead set
during the call to `init` from the 'segment_func' key-word argument.

The WorkerMapper class and related Worker and Task classes will be
provided as-is and are not really intended to be superclasses although
you are free to do so.

The only interfacing that facilitates their usage is that the
simulation manager will pass a required keyword argument 'num_workers'
to the call to `init`.

See the simulation manager module to see what fields are passed to the
mappers. These will likely not be removed in the future, although more
may be added.

"""



"""This module here is part of the RandomWalk object that implements
the distance metric for the RandomWalk walk system. This distance
metric is a scaled version of the Manhattan Norm.

"""

import logging

import numpy as np

from wepy.resampling.distances.distance import Distance


class RandomWalkDistance(Distance):
    """A class to implement the RandomWalkDistance metric for measuring
    differences between walker states. This is a normalized Manhattan
    distance measured between the difference in positions of the walkers.

    """

    def __init__(self):
        """Construct a RandomWalkDistance metric."""
        pass


    def image(self, state):
        """Transform a state into a random walk image.

        A random walk image is just the position of a walker in the
        N-dimensional space.

        Parameters
        ----------

        state : object implementing WalkerState
            A walker state object with positions in a numpy array
            of shape (N), where N is the the dimension of the random
            walk system.

        Returns
        -------

        randomwalk_image : array of floats of shape (N)
            The positions of a walker in the N-dimensional space.

        """
        return state['positions']

    def image_distance(self, image_a, image_b):
        """Compute the distance between the image of the two walkers.

        Parameters
        ----------

        image_a : array of float of shape (1, N)
            Position of the first walker's state.

        image_b:  array of float of shape (1, N)
            Position of the second walker's state.

       Returns
        -------

        distance: float
            The normalized Manhattan distance.

        """
        return np.average(np.abs(image_a - image_b))



from copy import copy, deepcopy
import sqlite3
from hashlib import md5
import time
import os
import os.path as osp
from base64 import b64encode, b64decode
from zlib import compress, decompress
import itertools as it
import logging

# instead of pickle we use dill, so we can save dynamically defined
# classes
import dill

from wepy.sim_manager import Manager
from wepy.orchestration.configuration import Configuration
from wepy.orchestration.snapshot import SimApparatus, SimSnapshot

from wepy.util.kv import KV, SQLITE3_INMEMORY_URI, gen_uri


class OrchestratorError(Exception):
    """ """
    pass



class Orchestrator():
    """ """

    # we freeze the pickle protocol for making hashes, because we care
    # more about stability than efficiency of newer versions
    HASH_PICKLE_PROTOCOL = 3

    DEFAULT_WORKDIR = Configuration.DEFAULT_WORKDIR
    DEFAULT_CONFIG_NAME = Configuration.DEFAULT_CONFIG_NAME
    DEFAULT_NARRATION = Configuration.DEFAULT_NARRATION
    DEFAULT_MODE = Configuration.DEFAULT_MODE

    DEFAULT_CHECKPOINT_FILENAME = "checkpoint.orch.sqlite"
    ORCH_FILENAME_TEMPLATE = "{config}{narration}.orch.sqlite"

    # the default way to oepn up the whole parent database
    DEFAULT_ORCHESTRATION_MODE = 'x'

    # mode to open the individual kv stores on the parent database
    KV_MODE = 'r+'

    # default timeout for connecting to a database
    SQLITE3_DEFAULT_TIMEOUT = 5

    # the fields to return (and their order) as a record for a run
    # query
    RUN_SELECT_FIELDS = ('last_cycle_idx', 'config_hash')

    def __init__(self, orch_path=None,
                 mode='x',
                 append_only=False,
    ):

        self._mode = mode
        self._append_only = append_only

        # handle the path and convert to a proper URI for the database
        # given the path and the mode
        self._db_uri = gen_uri(orch_path, mode)

        # run table: start_hash, end_hash, num_cycles, configuration_id

        # get a raw connection to the database
        self._db = sqlite3.connect(self.db_uri, uri=True,
                                   timeout=self.SQLITE3_DEFAULT_TIMEOUT)
        self._closed = False

        # set isolation level to autocommit
        self._db.isolation_level = None

        # we can use read_uncommited only in append_only mode (no
        # updates) because you never have to worry about dirty reads
        # since you can't update
        if self.append_only:
            self._db.execute("PRAGMA read_uncommited=1")

        # we make a table for the run data, if it doesn't already
        # exist
        c = self._db.cursor().execute(self.create_run_table_query)


        # initialize or open each of the separate KV-stores (tables in
        # the same SQLite3 database)

        # change the mode for the KV stores since we already created the database

        # metadata: default init walkers, default apparatus, default
        # configuration
        self.metadata_kv = KV(db_url=self.db_uri,
                              table='meta',
                              mode='a',
                              value_types=None,
                              append_only=self.append_only)



        # snapshots
        self.snapshot_kv = KV(db_url=self.db_uri,
                              table='snapshots',
                              primary_key='snaphash',
                              value_name='snapshot',
                              mode='a',
                              append_only=self.append_only)

        # configurations
        self.configuration_kv = KV(db_url=self.db_uri,
                                   table='configurations',
                                   primary_key='config_hash',
                                   value_name='config',
                                   mode='a',
                                   append_only=self.append_only)

    @property
    def mode(self):
        return self._mode

    @property
    def append_only(self):
        return self._append_only

    def close(self):

        if self._closed == True:
            raise IOError("The database connection is already closed")

        else:
            # close all the connections
            self.metadata_kv.close()
            self.configuration_kv.close()
            self.snapshot_kv.close()
            self._db.close()
            self._closed = True

    @property
    def db_uri(self):
        return self._db_uri

    @property
    def orch_path(self):

        # if it is not an in-memory database we parse off the path and
        # return that
        if self.db_uri == SQLITE3_INMEMORY_URI:
            return None
        else:

            # URIs have the following form: protocol:url?query
            # destructure the URI
            _, tail = self.db_uri.split(':')

            if len(tail.split('?')) > 1:
                url, _ = tail.split('?')
            else:
                url = tail

            return url


    @classmethod
    def serialize(cls, snapshot):
        """Serialize a snapshot to a compressed, encoded, pickle string
        representation.

        Currently uses the dill module for pickling because the base
        pickle module is inadequate. However, it is mostly compatible
        and can be read natively with pickle but this usage is
        officially not supported. Instead use the deserialize_snapshot.

        Also compresses with default zlib compression and is encoded
        in base64.

        The object will always have a deepcopy performed on it so that
        all of the extraneous references to it are avoided since there
        is no (AFAIK) way to make sure all references to an object are
        deleted.

        NOTE: Perhaps there is a way and that should be done (and
        tested) to see if it provides stable pickles (i.e. pickles
        that always hash to the same value). To avoid the overhead of
        copying large objects.

        Parameters
        ----------
        snapshot : SimSnapshot object
            The snapshot of the simulation you want to serialize.

        Returns
        -------
        serial_str : str
            Serialized string of the snapshot object

        """

        serial_str = b64encode(
                        compress(
                            dill.dumps(
                                deepcopy(snapshot),
                                protocol=cls.HASH_PICKLE_PROTOCOL,
                                recurse=True)
                        )
                     )

        return serial_str


    # core methods for serializing python objects, used for snapshots,
    # apparatuses, configurations, and the initial walker list

    @classmethod
    def deserialize(cls, serial_str):
        """Deserialize an unencoded string snapshot to an object.

        Parameters
        ----------
        serial_str : str
            Serialized string of the snapshot object

        Returns
        -------
        snapshot : SimSnapshot object
            Simulation snapshot object

        """

        return dill.loads(decompress(b64decode(serial_str)))


    # defaults getters and setters
    def set_default_sim_apparatus(self, sim_apparatus):

        # serialize the apparatus and then set it
        serial_app = self.serialize(sim_apparatus)

        self.metadata_kv['default_sim_apparatus'] = serial_app

    def set_default_init_walkers(self, init_walkers):

        # serialize the apparatus and then set it
        serial_walkers = self.serialize(init_walkers)

        self.metadata_kv['default_init_walkers'] = serial_walkers

    def set_default_configuration(self, configuration):

        # serialize the apparatus and then set it
        serial_config = self.serialize(configuration)

        config_hash = self.hash_snapshot(serial_config)

        self.metadata_kv['default_configuration_hash'] = config_hash

        self.configuration_kv[config_hash] = serial_config

    def set_default_snapshot(self, snapshot):

        snaphash = self.add_snapshot(snapshot)

        # then save the hash in the metadata
        self.metadata_kv['default_snapshot_hash'] = snaphash

        return snaphash

    def gen_default_snapshot(self):

        # generate the snapshot
        sim_start_hash = self.gen_start_snapshot(self.get_default_init_walkers())

        # then save the hash in the metadata
        self.metadata_kv['default_snapshot_hash'] = sim_start_hash

        return sim_start_hash


    def get_default_sim_apparatus(self):

        return self.deserialize(self.metadata_kv['default_sim_apparatus'])

    def get_default_init_walkers(self):

        return self.deserialize(self.metadata_kv['default_init_walkers'])

    def get_default_configuration(self):

        config_hash = self.metadata_kv['default_configuration_hash']

        return self.get_configuration(config_hash)

    def get_default_configuration_hash(self):

        return self.metadata_kv['default_configuration_hash']


    def get_default_snapshot(self):

        start_hash = self.metadata_kv['default_snapshot_hash']

        return self.get_snapshot(start_hash)

    def get_default_snapshot_hash(self):

        return self.metadata_kv['default_snapshot_hash']


    @classmethod
    def hash_snapshot(cls, serial_str):
        """

        Parameters
        ----------
        serial_str :
            

        Returns
        -------

        """
        return md5(serial_str).hexdigest()


    def get_snapshot(self, snapshot_hash):
        """Returns a copy of a snapshot.

        Parameters
        ----------
        snapshot_hash :
            

        Returns
        -------

        """

        return self.deserialize(self.snapshot_kv[snapshot_hash])


    def get_configuration(self, config_hash):
        """Returns a copy of a snapshot.

        Parameters
        ----------
        config_hash :
            

        Returns
        -------

        """

        return self.deserialize(self.configuration_kv[config_hash])


    @property
    def snapshot_hashes(self):
        """ """

        # iterate over the snapshot kv
        return list(self.snapshot_kv.keys())

    @property
    def configuration_hashes(self):
        """ """

        # iterate over the snapshot kv
        return list(self.configuration_kv.keys())


    def add_snapshot(self, snapshot):
        """

        Parameters
        ----------
        snapshot :

        Returns
        -------

        """

        # serialize the snapshot using the protocol for doing so
        serialized_snapshot = self.serialize(snapshot)

        # get the hash of the snapshot
        snaphash = self.hash_snapshot(serialized_snapshot)

        # check that the hash is not already in the snapshots
        if any([True if snaphash == md5 else False for md5 in self.snapshot_hashes]):

            # just skip the rest of the function and return the hash
            return snaphash

        # save the snapshot in the KV store
        self.snapshot_kv[snaphash] = serialized_snapshot

        return snaphash

    def add_serial_snapshot(self, serial_snapshot):

        # get the hash of the snapshot
        snaphash = self.hash_snapshot(serial_snapshot)

        # check that the hash is not already in the snapshots
        if any([True if snaphash == md5 else False for md5 in self.snapshot_hashes]):

            # just skip the rest of the function and return the hash
            return snaphash

        # save the snapshot in the KV store
        self.snapshot_kv[snaphash] = serial_snapshot

        return snaphash

    def gen_start_snapshot(self, init_walkers):
        """

        Parameters
        ----------
        init_walkers :
            

        Returns
        -------

        """

        # make a SimSnapshot object using the initial walkers and
        start_snapshot = SimSnapshot(init_walkers, self.get_default_sim_apparatus())

        # save the snapshot, and generate its hash
        sim_start_md5 = self.add_snapshot(start_snapshot)

        return sim_start_md5

    @property
    def default_snapshot_hash(self):
        """ """
        return self.metadata_kv['default_snapshot_hash']

    @property
    def default_snapshot(self):
        """ """
        return self.get_snapshot(self.default_snapshot_hash)

    def snapshot_registered(self, snapshot):
        """Check whether a snapshot is already in the database, based on the
        hash of it.

        This serializes the snapshot so may be slow.

        Parameters
        ----------
        snapshot : SimSnapshot object
            The snapshot object you want to query for.

        Returns
        -------

        """

        # serialize and hash the snapshot
        snaphash = self.hash_snapshot(self.serialize(snapshot))

        # then check it
        return self.snapshot_hash_registered(snaphash)

    def snapshot_hash_registered(self, snapshot_hash):
        """Check whether a snapshot hash is already in the database.

        Parameters
        ----------
        snapshot_hash : str
            The string hash of the snapshot.

        Returns
        -------

        """

        if any([True if snapshot_hash == h else False for h in self.snapshot_hashes]):
            return True
        else:
            return False


    def configuration_hash_registered(self, config_hash):
        """Check whether a snapshot hash is already in the database.

        Parameters
        ----------
        snapshot_hash : str
            The string hash of the snapshot.

        Returns
        -------

        """

        if any([True if config_hash == h else False for h in self.configuration_hashes]):
            return True
        else:
            return False


    ### run methods

    def add_configuration(self, configuration):

        serialized_config = self.serialize(configuration)

        config_hash = self.hash_snapshot(serialized_config)

        # check that the hash is not already in the snapshots
        if any([True if config_hash == md5 else False for md5 in self.configuration_hashes]):

            # just skip the rest of the function and return the hash
            return config_hash

        # save the snapshot in the KV store
        self.configuration_kv[config_hash] = serialized_config

        return config_hash

    def add_serial_configuration(self, serial_configuration):

        # get the hash of the configuration
        snaphash = self.hash_snapshot(serial_configuration)

        # check that the hash is not already in the configurations
        if any([True if snaphash == md5 else False for md5 in self.configuration_hashes]):

            # just skip the rest of the function and return the hash
            return snaphash

        # save the configuration in the KV store
        self.configuration_kv[snaphash] = serial_configuration

        return snaphash


    @property
    def create_run_table_query(self):
        create_run_table_query = """
            CREATE TABLE IF NOT EXISTS runs
             (start_hash TEXT NOT NULL,
             end_hash TEXT NOT NULL,
             config_hash NOT NULL,
             last_cycle_idx INTEGER NOT NULL,
             PRIMARY KEY (start_hash, end_hash))

        """

        return create_run_table_query

    @property
    def add_run_record_query(self):

        add_run_row_query = """
        INSERT INTO runs (start_hash, end_hash, config_hash, last_cycle_idx)
        VALUES (?, ?, ?, ?)
        """

        return add_run_row_query

    @property
    def update_run_record_query(self):

        q = """
        UPDATE runs
        SET config_hash = ?,
            last_cycle_idx = ?
        WHERE start_hash=? AND end_hash=?
        """

        return q

    @property
    def delete_run_record_query(self):

        q = """
        DELETE FROM runs
        WHERE start_hash=? AND end_hash=?
        """

        return q


    def _add_run_record(self, start_hash, end_hash, configuration_hash, cycle_idx):

        params = (start_hash, end_hash, configuration_hash, cycle_idx)

        # do it as a transaction
        c = self._db.cursor()

        # run the insert
        c.execute(self.add_run_record_query, params)

    def _delete_run_record(self, start_hash, end_hash):

        params = (start_hash, end_hash)

        cursor = self._db.cursor()

        cursor.execute(self.delete_run_record_query, params)

    def _update_run_record(self, start_hash, end_hash, new_config_hash, new_last_cycle_idx):

        params = (new_config_hash, new_last_cycle_idx, start_hash, end_hash)

        # do it as a transaction
        c = self._db.cursor()

        # run the update
        c.execute(self.update_run_record_query, params)



    def register_run(self, start_hash, end_hash, config_hash, cycle_idx):
        """

        Parameters
        ----------
        start_hash :
            
        end_hash :
            
        config_hash :
            
        cycle_idx : int
            The cycle of the simulation run the checkpoint was generated for.

        Returns
        -------

        """

        # check that the hashes are for snapshots in the orchestrator
        # if one is not registered raise an error
        if not self.snapshot_hash_registered(start_hash):
            raise OrchestratorError(
                "snapshot start_hash {} is not registered with the orchestrator".format(
                start_hash))

        if not self.snapshot_hash_registered(end_hash):
            raise OrchestratorError(
                "snapshot end_hash {} is not registered with the orchestrator".format(
                end_hash))

        if not self.configuration_hash_registered(config_hash):
            raise OrchestratorError(
                "config hash {} is not registered with the orchestrator".format(
                config_hash))


        # save the configuration and get it's id

        self._add_run_record(start_hash, end_hash, config_hash, cycle_idx)

    def get_run_records(self):

        get_run_record_query = """
        SELECT *
        FROM runs
        """.format(fields=', '.join(self.RUN_SELECT_FIELDS))

        cursor = self._db.cursor()
        cursor.execute(get_run_record_query)
        records = cursor.fetchall()

        return records

    def get_run_record(self, start_hash, end_hash):

        get_run_record_query = """
        SELECT {fields}
        FROM runs
        WHERE start_hash=? AND end_hash=?
        """.format(fields=', '.join(self.RUN_SELECT_FIELDS))

        params = (start_hash, end_hash)

        cursor = self._db.cursor()
        cursor.execute(get_run_record_query, params)
        record = cursor.fetchone()

        return record

    def run_last_cycle_idx(self, start_hash, end_hash):

        record = self.get_run_record(start_hash, end_hash)

        last_cycle_idx = record[self.RUN_SELECT_FIELDS.index('last_cycle_idx')]

        return last_cycle_idx


    def run_configuration(self, start_hash, end_hash):

        record = self.get_run_record(start_hash, end_hash)

        config_hash = record[self.RUN_SELECT_FIELDS.index('config_hash')]

        # get the configuration object and deserialize it
        return self.deserialize(self.configuration_kv[config_hash])

    def run_configuration_hash(self, start_hash, end_hash):

        record = self.get_run_record(start_hash, end_hash)

        config_hash = record[self.RUN_SELECT_FIELDS.index('config_hash')]

        return config_hash

    def run_hashes(self):

        return [(rec[0], rec[1]) for rec in self.get_run_records()]

    def run_continues(self, start_hash, end_hash):
        """Given a start hash and end hash for a run, find the run that this
        continues.

        Parameters
        ----------
        start_hash :
            
        end_hash :
            

        Returns
        -------
        run_id

        """

        # loop through the runs in this orchestrator until we find one
        # where the start_hash matches the end hash
        runs = self.run_hashes()
        run_idx = 0
        while True:

            run_start_hash, run_end_hash = runs[run_idx]

            # if the start hash of the queried run is the same as the
            # end hash for this run we have found it
            if start_hash == run_end_hash:

                return (run_start_hash, run_end_hash)

            run_idx += 1

            # if the index is over the number of runs we quit and
            # return None as no match
            if run_idx >= len(runs):
                return None


    def _init_checkpoint_db(self, start_hash, configuration, checkpoint_dir, mode='x'):

        logging.debug("Initializing checkpoint orch database")

        # make the checkpoint with the default filename at the checkpoint directory
        checkpoint_path = osp.join(checkpoint_dir, self.DEFAULT_CHECKPOINT_FILENAME)

        # create a new database in the mode specified
        logging.debug("Creating checkpoint database")
        checkpoint_orch = Orchestrator(checkpoint_path, mode=mode)

        # add the starting snapshot, bypassing the serialization stuff
        logging.debug("Setting the starting snapshot")
        checkpoint_orch.snapshot_kv[start_hash] = self.snapshot_kv[start_hash]

        # if we have a new configuration at runtime serialize and
        # hash it
        serialized_config = self.serialize(configuration)
        config_hash = self.hash_snapshot(serialized_config)

        # save the configuration as well
        checkpoint_orch.configuration_kv[config_hash] = serialized_config

        checkpoint_orch.close()
        logging.debug("closing connection to checkpoint database")

        return checkpoint_path, config_hash


    def _save_checkpoint(self, checkpoint_snapshot, config_hash,
                         checkpoint_db_path, cycle_idx,
                         ):
        """

        Parameters
        ----------
        checkpoint_snapshot :
            
        config_hash :
            
        checkpoint_db_path :
            
        mode :
             (Default value = 'wb')

        Returns
        -------

        """

        # orchestrator wrapper to the db
        logging.debug("Opening the checkpoint orch database")
        checkpoint_orch = Orchestrator(checkpoint_db_path, mode='r+')

        # connection to the db
        cursor = checkpoint_orch._db.cursor()

        # we replicate the code for adding the snapshot here because
        # we want it to occur transactionally the delete and add

        # serialize the snapshot using the protocol for doing so
        serialized_snapshot = self.serialize(checkpoint_snapshot)

        # get the hash of the snapshot
        snaphash = self.hash_snapshot(serialized_snapshot)

        # the queries for deleting and inserting the new run record
        delete_query = """
        DELETE FROM runs
        WHERE start_hash=?
          AND end_hash=?
        """

        insert_query = """
        INSERT INTO runs (start_hash, end_hash, config_hash, last_cycle_idx)
        VALUES (?, ?, ?, ?)
        """

        # if there are any runs in the checkpoint orch remove the
        # final snapshot
        delete_params = None
        if len(checkpoint_orch.run_hashes()) > 0:
            start_hash, old_checkpoint_hash = checkpoint_orch.run_hashes()[0]

            delete_params = (start_hash, old_checkpoint_hash)
        else:
            start_hash = list(checkpoint_orch.snapshot_kv.keys())[0]

        # the config should already be in the orchestrator db
        insert_params = (start_hash, snaphash, config_hash, cycle_idx)


        # start this whole process as a transaction so we don't get
        # something weird in between
        logging.debug("Starting transaction for updating run table in checkpoint")
        cursor.execute("BEGIN TRANSACTION")

        # add the new one, using a special method for setting inside
        # of a transaction
        logging.debug("setting the new checkpoint snapshot into the KV")
        cursor = checkpoint_orch.snapshot_kv.set_in_tx(cursor, snaphash, serialized_snapshot)
        logging.debug("finished")

        # if we need to delete the old end of the run snapshot and the
        # run record for it
        if delete_params is not None:

            logging.debug("Old run record needs to be removed")

            # remove the old run from the run table
            logging.debug("Deleting the old run record")
            cursor.execute(delete_query, delete_params)
            logging.debug("finished")

        # register the new run in the run table
        logging.debug("Inserting the new run record")
        cursor.execute(insert_query, insert_params)
        logging.debug("finished")

        # end the transaction
        logging.debug("Finishing transaction")
        cursor.execute("COMMIT")
        logging.debug("Transaction committed")

        # we do the removal of the old snapshot outside of the
        # transaction since it is slow and can cause timeouts to
        # occur. Furthermore, it is okay if it is in the checkpoint as
        # the run record is what matters as long as the new checkpoint
        # is there.

        # delete the old snapshot if we need to
        if delete_params is not None:
            logging.debug("Deleting the old snapshot")
            del checkpoint_orch.snapshot_kv[old_checkpoint_hash]
            logging.debug("finished")


        checkpoint_orch.close()
        logging.debug("closed the checkpoint orch connection")


    @staticmethod
    def gen_sim_manager(start_snapshot, configuration):
        """

        Parameters
        ----------
        start_snapshot :
            
        configuration :
            

        Returns
        -------

        """

        # construct the sim manager, in a wepy specific way
        sim_manager = Manager(start_snapshot.walkers,
                              runner=start_snapshot.apparatus.filters[0],
                              boundary_conditions=start_snapshot.apparatus.filters[1],
                              resampler=start_snapshot.apparatus.filters[2],
                              # configuration options
                              work_mapper=configuration.work_mapper,
                              reporters=configuration.reporters,
                              sim_monitor=configuration.monitor,
        )

        return sim_manager


    def run_snapshot_by_time(self, start_hash, run_time, n_steps,
                             checkpoint_freq=None,
                             checkpoint_dir=None,
                             configuration=None,
                             configuration_hash=None,
                             checkpoint_mode='x'):
        """For a finished run continue it but resetting all the state of the
        resampler and boundary conditions

        Parameters
        ----------
        start_hash :
            
        run_time :
            
        n_steps :
            
        checkpoint_freq :
             (Default value = None)
        checkpoint_dir :
             (Default value = None)
        configuration :
             (Default value = None)
        configuration_hash :
             (Default value = None)
        checkpoint_mode :
             (Default value = None)

        Returns
        -------

        """


        # you must have a checkpoint dir if you ask for a checkpoint
        # frequency
        if checkpoint_freq is not None and checkpoint_dir is None:
            raise ValueError("Must provide a directory for the checkpoint file "
                             "is a frequency is specified")


        if configuration_hash is not None and configuration is not None:
            raise ValueError("Cannot specify both a hash of an existing configuration"
                             "and provide a runtime configuration")

        # if no configuration was specified we use the default one, oth
        elif (configuration is None) and (configuration_hash is None):
            configuration = self.get_default_configuration()

        # if a configuration hash was given only then we retrieve that
        # configuration since we must pass configurations to the
        # checkpoint DB initialization
        elif configuration_hash is not None:
            configuration = self.configuration_kv[configuration_hash]


        # check that the directory for checkpoints exists, and create
        # it if it doesn't and isn't already created
        if checkpoint_dir is not None:
            checkpoint_dir = osp.realpath(checkpoint_dir)
            os.makedirs(checkpoint_dir, exist_ok=True)


        # if the checkpoint dir is not specified don't create a
        # checkpoint db orch
        checkpoint_db_path = None
        if checkpoint_dir is not None:
            logging.debug("Initialization of checkpoint database is requested")
            checkpoint_db_path, configuration_hash = self._init_checkpoint_db(start_hash,
                                                                              configuration,
                                                                              checkpoint_dir,
                                                                              mode=checkpoint_mode)
            logging.debug("finished initializing checkpoint database")


        # get the snapshot and the configuration to use for the sim_manager
        start_snapshot = self.get_snapshot(start_hash)

        # generate the simulation manager given the snapshot and the
        # configuration
        sim_manager = self.gen_sim_manager(start_snapshot, configuration)

        # handle and process the optional arguments for running simulation
        if 'runner' in configuration.apparatus_opts:
            runner_opts = configuration.apparatus_opts['runner']
        else:
            runner_opts = None

        # run the init subroutine for the simulation manager
        logging.debug("Running sim_manager.init")
        sim_manager.init()

        # run each cycle manually creating checkpoints when necessary
        logging.debug("Starting run loop")
        walkers = sim_manager.init_walkers
        cycle_idx = 0
        start_time = time.time()
        while time.time() - start_time < run_time:

            logging.debug("Running cycle {}".format(cycle_idx))
            # run the cycle
            walkers, filters = sim_manager.run_cycle(
                walkers,
                n_steps,
                cycle_idx,
                runner_opts=runner_opts,
            )

            # check to see if a checkpoint is necessary
            if (checkpoint_freq is not None):
                if (cycle_idx % checkpoint_freq == 0):
                    logging.debug("Checkpoint is required for this cycle")

                    # make the checkpoint snapshot
                    logging.debug("Generating the simulation snapshot")
                    checkpoint_snapshot = SimSnapshot(walkers, SimApparatus(filters))

                    # save the checkpoint (however that is implemented)
                    logging.debug("saving the checkpoint to the database")
                    self._save_checkpoint(checkpoint_snapshot,
                                          configuration_hash,
                                          checkpoint_db_path,
                                          cycle_idx)
                    logging.debug("finished saving the checkpoint to the database")

            # increase the cycle index for the next cycle
            cycle_idx += 1

        logging.debug("Finished the run cycle")

        # the cycle index was set for the next cycle which didn't run
        # so we decrement it
        last_cycle_idx = cycle_idx - 1

        logging.debug("Running sim_manager.cleanup")
        # run the cleanup subroutine
        sim_manager.cleanup()

        # run the segment given the sim manager and run parameters
        end_snapshot = SimSnapshot(walkers, SimApparatus(filters))

        logging.debug("Run finished")
        # return the things necessary for saving to the checkpoint if
        # that is what is wanted later on
        return end_snapshot, configuration_hash, checkpoint_db_path, last_cycle_idx

    def orchestrate_snapshot_run_by_time(self, snapshot_hash, run_time, n_steps,
                                         checkpoint_freq=None,
                                         checkpoint_dir=None,
                                         orchestrator_path=None,
                                         configuration=None,
                                         # these can reparametrize the paths
                                         # for both the orchestrator produced
                                         # files as well as the configuration
                                         work_dir=None,
                                         config_name=None,
                                         narration=None,
                                         mode=None,
                                         # extra kwargs will be passed to the
                                         # configuration.reparametrize method
                                         **kwargs):
        """

        Parameters
        ----------
        snapshot_hash :
            
        run_time :
            
        n_steps :
            
        checkpoint_freq :
             (Default value = None)
        checkpoint_dir :
             (Default value = None)
        orchestrator_path :
             (Default value = None)
        configuration :
             (Default value = None)
        # these can reparametrize the paths# for both the orchestrator produced# files as well as the configurationwork_dir :
             (Default value = None)
        config_name :
             (Default value = None)
        narration :
             (Default value = None)
        mode :
             (Default value = None)
        # extra kwargs will be passed to the# configuration.reparametrize method**kwargs :
            

        Returns
        -------

        """


        # for writing the orchestration files we set the default mode
        # if mode is not given
        if mode is None:
            # the orchestrator mode is used for pickling the
            # orchestrator and so must be in bytes mode
            orch_mode = self.DEFAULT_ORCHESTRATION_MODE

        # there are two possible uses for the path reparametrizations:
        # the configuration and the orchestrator file paths. If both
        # of those are explicitly specified by passing in the whole
        # configuration object or both of checkpoint_dir,
        # orchestrator_path then those reparametrization kwargs will
        # not be used. As this is likely not the intention of the user
        # we will raise an error. If there is even one use for them no
        # error will be raised.

        # first check if any reparametrizations were even requested
        parametrizations_requested = (True if work_dir is not None else False,
                                      True if config_name is not None else False,
                                      True if narration is not None else False,
                                      True if mode is not None else False,)

        # check if there are any available targets for reparametrization
        reparametrization_targets = (True if configuration is None else False,
                                     True if checkpoint_dir is None else False,
                                     True if orchestrator_path is None else False)

        # if paramatrizations were requested and there are no targets
        # we need to raise an error
        if any(parametrizations_requested) and not any(reparametrization_targets):

            raise OrchestratorError("Reparametrizations were requested but none are possible,"
                                    " due to all possible targets being already explicitly given")


        # if any paths were not given and no defaults for path
        # parameters we want to fill in the defaults for them. This
        # will also fill in any missing parametrizations with defaults

        # we do this by just setting the path parameters if they
        # aren't set, then later the parametrization targets will be
        # tested for if they have been set or not, and if they haven't
        # then these will be used to generate paths for them.
        if work_dir is None:
            work_dir = self.DEFAULT_WORKDIR
        if config_name is None:
            config_name = self.DEFAULT_CONFIG_NAME
        if narration is None:
            narration = self.DEFAULT_NARRATION
        if mode is None:
           mode = self.DEFAULT_MODE

        # if no configuration was specified use the default one
        if configuration is None:
            configuration = self.get_default_configuration()

            # reparametrize the configuration with the given path
            # parameters and anything else in kwargs. If they are none
            # this will have no effect anyhow
            logging.debug("Reparametrizing the configuration")
            configuration = configuration.reparametrize(work_dir=work_dir,
                                                        config_name=config_name,
                                                        narration=narration,
                                                        mode=mode,
                                                        **kwargs)

        # make parametric paths for the checkpoint directory and the
        # orchestrator pickle to be made, unless they are explicitly given

        if checkpoint_dir is None:

            # the checkpoint directory will be in the work dir
            logging.debug("checkpoint directory defaulted to the work_dir")
            checkpoint_dir = work_dir


        logging.debug("In the orchestrate run, calling to run_snapshot by time")
        # then actually run the simulation with checkpointing. This
        # returns the end snapshot and doesn't write out anything to
        # orchestrators other than the checkpointing
        (end_snapshot, configuration_hash, checkpoint_db_path, last_cycle_idx) =\
            self.run_snapshot_by_time(snapshot_hash, run_time, n_steps,
                                            checkpoint_freq=checkpoint_freq,
                                            checkpoint_dir=checkpoint_dir,
                                            configuration=configuration,
                                            checkpoint_mode=orch_mode)

        logging.debug("Finished running snapshot by time")

        # if the last cycle in the run was a checkpoint skip this step
        # of saving a checkpoint
        do_final_checkpoint = True

        # make sure the checkpoint_freq is defined before testing it
        if checkpoint_freq is not None:
            if checkpoint_freq % last_cycle_idx == 0:
                logging.debug("Last cycle saved a checkpoint, no need to save one")
                do_final_checkpoint = False

        if do_final_checkpoint:

            logging.debug("Saving a final checkpoint for the end of the run")
            # now that it is finished we save the final snapshot to the
            # checkpoint file. This is done transactionally using the
            # SQLite transaction functionality (either succeeds or doesn't
            # happen) that way we don't have worry about data integrity
            # loss. Here we also don't have to worry about other processes
            # interacting with the checkpoint which makes it isolated.
            self._save_checkpoint(end_snapshot, configuration_hash,
                                  checkpoint_db_path, last_cycle_idx)
            logging.debug("Finished saving the final checkpoint for the run")

        # then return the final orchestrator
        logging.debug("Getting a connection to that orch to retun")
        checkpoint_orch = Orchestrator(checkpoint_db_path,
                                       mode='r+',
                                       append_only=True)

        return checkpoint_orch




def reconcile_orchestrators(host_path, *orchestrator_paths):
    """

    Parameters
    ----------
    template_orchestrator :
        
    *orchestrators :
        

    Returns
    -------

    """

    if not osp.exists(host_path):
        assert len(orchestrator_paths) > 1, \
            "If the host path is a new orchestrator, must give at least 2 orchestrators to merge."

    # open the host orchestrator at the location which will have all
    # of the new things put into it from the other orchestrators. If
    # it doesn't already exist it will be created otherwise open
    # read-write.
    new_orch = Orchestrator(orch_path=host_path,
                            mode='a',
                            append_only=True)

    # TODO deprecate, if there is no defaults we can't set them since
    # the mode is append only, we don't really care about these so
    # don't set them, otherwise do some mode logic to figure this out
    # and open in write mode and set defaults, then change to append
    # only

    # # if this is an existing orchestrator copy the default
    # # sim_apparatus and init_walkers
    # try:
    #     default_app = new_orch.get_default_sim_apparatus()
    # except KeyError:
    #     # no default apparatus, that is okay
    #     pass
    # else:
    #     # set it
    #     new_orch.set_default_sim_apparatus(default_app)

    # # same for the initial walkers
    # try:
    #     default_walkers = new_orch.get_default_init_walkers()
    # except KeyError:
    #     # no default apparatus, that is okay
    #     pass
    # else:
    #     # set it
    #     new_orch.set_default_sim_apparatus(default_walkers)


    for orch_path in orchestrator_paths:

        # open it in read-write fail if doesn't exist
        orch = Orchestrator(orch_path=orch_path,
                            mode='r+',
                            append_only=True)

        # add in all snapshots from each orchestrator, by the hash not the
        # snapshots themselves, we trust they are correct
        for snaphash in orch.snapshot_hashes:

            # check that the hash is not already in the snapshots
            if any([True if snaphash == md5 else False for md5 in new_orch.snapshot_hashes]):

                # skip it and move on
                continue

            # if it is not copy it over without deserializing
            new_orch.snapshot_kv[snaphash] = orch.snapshot_kv[snaphash]

        # add in the configurations for the runs from each
        # orchestrator, by the hash not the snapshots themselves, we
        # trust they are correct
        for run_id in orch.run_hashes():

            config_hash = orch.run_configuration_hash(*run_id)

            # check that the hash is not already in the snapshots
            if any([True if config_hash == md5 else False for md5 in new_orch.configuration_hashes]):

                # skip it and move on
                continue

            # if it is not set it
            new_orch.configuration_kv[config_hash] = orch.configuration_kv[config_hash]

        # concatenate the run table with an SQL union from an attached
        # database

        attached_table_name = "other"

        # query to attach the foreign database
        attach_query = """
        ATTACH '{}' AS {}
        """.format(orch_path, attached_table_name)

        # query to update the runs tabel with new unique runs
        union_query = """
        INSERT INTO runs
        SELECT * FROM (
        SELECT * FROM {}.runs
        EXCEPT
        SELECT * FROM runs
        )
        """.format(attached_table_name)

        # query to detach the table
        detach_query = """
        DETACH {}
        """.format(attached_table_name)

        # then run the queries

        cursor = new_orch._db.cursor()
        try:
            cursor.execute('BEGIN TRANSACTION')
            cursor.execute(attach_query)
            cursor.execute(union_query)
            cursor.execute('COMMIT')
            cursor.execute(detach_query)
        except:
            cursor.execute('COMMIT')
            import pdb; pdb.set_trace()
            cursor.execute("SELECT * FROM (SELECT * FROM other.runs EXCEPT SELECT * FROM runs)")
            recs = cursor.fetchall()

    return new_orch



"""Abstract base class for Decision classes.

See the NoDecision class and others in this module for examples.

To create your own subclass of the Decision class you must customize
the following class constants:

- ENUM
- FIELDS
- SHAPES
- DTYPE
- RECORD_FIELDS

The 'ENUM' field should be a python 'Enum' class created by
subclassing and customizing 'Enum' in the normal pythonic way. The
elements of the 'Enum' are the actual decision choices, and their
numeric value is used for serialization.

The 'FIELDS' constant is a specification of the number of fields that
a decision record will have. All decision records should contain the
'decision_id' field which is the choice of decision. This class
implements that and should be used as shown in the examples.

In order that fields be serializable to different formats, we also
require that they be a numpy array datatype.

To support this we require the data shapes and data types for each
field. Elements of SHAPES and DTYPES should be of a format
recognizable by the numpy array constructor.

To this we allow the additional option of specifying SHAPES as the
python built-in name Ellipsis (AKA '...'). This will specify the shape
as a variable length 1-dimensional array.

The RECORD_FIELDS is used as a way to specify fields which are
amenable to placement in simplified summary tables, i.e. simple
non-compound values.

The only method that needs to be implemented in the Decision is
'action'.

This function actually implements the algorithm for taking actions on
the decisions and instructions and is called from the resampler to
perform them on the collection of walkers.

"""

from collections import namedtuple
from enum import Enum
from string import ascii_lowercase
import logging

import numpy as np

# ABC for the Decision class
class Decision(object):
    """Represents and provides methods for a set of decision values.

    """

    ENUM = None
    """The enumeration of the decision types. Maps them to integers."""

    DEFAULT_DECISION = None
    """The default decision to choose."""

    FIELDS = ('decision_id',)
    """The names of the fields that go into the decision record."""

    # suggestion for subclassing, FIELDS and others
    # FIELDS = super().FIELDS + ('target_idxs',)
    # etc.

    #  An Ellipsis instead of fields indicate there is a variable
    # number of fields.
    SHAPES = ((1,),)
    """Field data shapes."""

    DTYPES = (np.int,)
    """Field data types."""

    RECORD_FIELDS = ('decision_id',)
    """The fields that could be used in a reduced table-like representation."""

    ANCESTOR_DECISION_IDS = None
    """Specify the enum values where their walker state sample value is
    passed on in the next generation, i.e. after performing the action."""

    @classmethod
    def default_decision(cls):
        return cls.DEFAULT_DECISION

    @classmethod
    def field_names(cls):
        """Names of the decision record fields."""
        return cls.FIELDS

    @classmethod
    def field_shapes(cls):
        """Field data shapes."""
        return cls.SHAPES

    @classmethod
    def field_dtypes(cls):
        """Field data types."""
        return cls.DTYPES

    @classmethod
    def fields(cls):
        """Specs for each field.

        Returns
        -------

        fields : list of tuples
            Field specs each spec is of the form (name, shape, dtype).

        """
        return list(zip(cls.field_names(),
                   cls.field_shapes(),
                   cls.field_dtypes()))

    @classmethod
    def record_field_names(cls):
        """The fields that could be used in a reduced table-like representation."""
        return cls.RECORD_FIELDS

    @classmethod
    def enum_dict_by_name(cls):
        """Get the decision enumeration as a dict mapping name to integer."""
        if cls.ENUM is None:
            raise NotImplementedError

        d = {}
        for enum in cls.ENUM:
            d[enum.name] = enum.value
        return d

    @classmethod
    def enum_dict_by_value(cls):
        """Get the decision enumeration as a dict mapping integer to name."""

        if cls.ENUM is None:
            raise NotImplementedError

        d = {}
        for enum in cls.ENUM:
            d[enum.value] = enum
        return d

    @classmethod
    def enum_by_value(cls, enum_value):
        """Get the enum name for an enum_value.

        Parameters
        ----------
        enum_value : int

        Returns
        -------
        enum_name : enum

        """
        d = cls.enum_dict_by_value()
        return d[enum_value]

    @classmethod
    def enum_by_name(cls, enum_name):
        """Get the enum name for an enum_value.

        Parameters
        ----------
        enum_name : enum

        Returns
        -------
        enum_value : int

        """

        d = cls.enum_dict_by_name()
        return d[enum_name]


    @classmethod
    def record(cls, enum_value, **fields):
        """Generate a record for the enum_value and the other fields.

        Parameters
        ----------
        enum_value : int

        Returns
        -------
        rec : dict of str: value

        """

        assert enum_value in cls.enum_dict_by_value(), "value is not a valid Enumerated value"

        for field_key in fields.keys():
            assert field_key in cls.FIELDS, \
                "The field {} is not a field for that decision".format(field_key)
            assert field_key != 'decision_id', "'decision_id' cannot be an extra field"

        rec = {'decision_id' : enum_value}
        rec.update(fields)

        return rec

    @classmethod
    def action(cls, walkers, decisions):
        """Perform the instructions for a set of resampling records on
        walkers.

        The decisions are a collection of decision records which
        contain the decision value and the instruction values for a
        particular walker within its cohort (sample set).

        The collection is organized as a list of lists. The outer list
        corresponds to the steps of resampling within a cycle.

        The inner list is a list of decision records for a specific
        step of resampling, where the index of the decision record is
        the walker index.

        Parameters
        ----------
        walkers : list of Walker objects
            The walkers you want to perform the decision instructions on.

        decisions : list of list of decision records
            The decisions for each resampling step and their
            instructions to apply to the walkers.

        Returns
        -------

        resampled_walkers : list of Walker objects
            The resampled walkers.

        Raises
        ------
        NotImplementedError : abstract method

        """
        raise NotImplementedError

    @classmethod
    def parents(cls, step):
        """Given a step of resampling records (for a single resampling step)
        returns the parents of the children of this step.

        Parameters
        ----------
        step : list of decision records
            The decision records for a step of resampling for each walker.

        Returns
        -------
        walker_step_parents : list of int
            For each element, the index of it in the list corresponds
            to the child index and the value of the element is the
            index of it's parent before the decision action.

        """

        # initialize a list for the parents of this stages walkers
        step_parents = [None for i in range(len(step))]

        # the rest of the stages parents are based on the previous stage
        for parent_idx, parent_rec in enumerate(step):

            # if the decision is an ancestor then the instruction
            # values will be the children
            if parent_rec[0] in cls.ANCESTOR_DECISION_IDS:
                # the first value of the parent record is the target
                # idxs
                child_idxs = parent_rec[1]
                for child_idx in child_idxs:
                    step_parents[child_idx] = parent_idx

        return step_parents




class NothingDecisionEnum(Enum):
    """Enumeration of the decision values for doing nothing."""

    NOTHING = 0
    """Do nothing with the walker."""

class NoDecision(Decision):
    """Decision for a resampling process that does no resampling."""

    ENUM = NothingDecisionEnum
    DEFAULT_DECISION = ENUM.NOTHING

    FIELDS = Decision.FIELDS + ('target_idxs',)
    SHAPES = Decision.SHAPES + (Ellipsis,)
    DTYPES = Decision.DTYPES + (np.int,)

    RECORD_FIELDS = Decision.RECORD_FIELDS + ('target_idxs',)

    ANCESTOR_DECISION_IDS = (ENUM.NOTHING.value,)

    @classmethod
    def action(cls, walkers, decisions):

        # list for the modified walkers
        mod_walkers = [None for i in range(len(walkers))]
        # go through each decision and perform the decision
        # instructions
        for walker_idx, decision in enumerate(decisions):
            decision_value, instruction = decision
            if decision_value == cls.ENUM.NOTHING.value:

                # check to make sure a walker doesn't already exist
                # where you are going to put it
                if mod_walkers[instruction[0]] is not None:
                    raise ValueError(
                        "Multiple walkers assigned to position {}".format(instruction[0]))

                # put the walker in the position specified by the
                # instruction
                mod_walkers[instruction[0]] = walkers[walker_idx]


        return mod_walkers



"""Module that provides functionality for performing parallel
distributed calculations using the dask suite of tools.

Because these tools are to be used in tandem with each other here is an overview of how they fit together:


Dask provides parallelism over arbitrary (serializable) python
objects through the bag object. Because WepyHDF5 objects (or any
HDF5 dataset reference) are not serializable (because they contain
file handles to an underlying HDF5 file) we instead provide a
simple datastructure that provides the specifications for which
dataset to read from and what piece of the data is to be read
in. These the 'items' (which are returned by this function) are a
simple tuple containing:

- file path to WepyHDF5 file
- run index
- trajectory index
- frame indices
- field names/paths

This is everything we need to call the `get_traj_field` method and
generate a traj fields, which can then be operated on by various
other functions, e.g. `traj_fields_to_mdtraj` or any function that
could be passed to `WepyHDF5.compute_observable`.

This function does not directly create a dask bag object, but can
easily be done, via
`dask.bag.from_sequence(traj_fields_bag_items(*args))`. After
which we only need to actually load the data into memory using the
other helper function `load_frames_fields` which takes as a single
argument a single 'item' from this functions output.

I.e.:
`dask.bag.from_sequence(traj_fields_bag_items(*args)).map(load_frames_fields)`

This will create a dask bag which can then be used to perform
various distributed, parallel operations on it, such as
`dask.bag.map`. Think of it as a distributed version of
`WepyHDF5.compute_observable` (except we can't save directly to
the WepyHDF5 in the same invocation).

For example:

```
import dask
from wepy.util.mdtraj import traj_fields_to_mdtraj
from wepy.analysis.distributed import traj_fields_chunk_items, load_traj_chunks

bag = dask.bag.from_sequence(
         traj_fields_bag_items(
                'results.wepy.h5',
                ['positions', 'box_vectors'],
                chunk_size=500
                              )
                            )

# load the trajectory fields "chunks" into distributed memory
a = bag.map(load_traj_chunks)
# convert our "trajectory fields" chunks to an mdtraj Trajectory object
b = bag.map(traj_fields_to_mdtraj)
# compute the solvent accessible surface area of the trajectory chunks
c = bag.map(mdj.shrake_rupley)
results = b.compute()

```

The intermediate values a, b, and c are just place holders for
delayed operations since, dask is 'lazy' in the sense that it
doesn't actually do any of the work in the 'map' calls until you
call 'compute'.

The choice of the chunk size is also of high importance for getting
efficient calculations. To choose the chunk size appropriately
consider the resources and parallelism available to each of your
workers and the amount of data that you have.

The smaller the chunk, the more tasks that will be generated for
dask. Each task has an intrinsic overhead associated with it, may
require serialization and communication if operating in separate
processes or hosts, and furthermore needs to be scheduled for
execution by dask. An excessively large number of tasks will cause the
scheduler to grind to a halt so we want to increase the chunk size to
large enough that the scheduler can handle it and that communication
doesn't become more expensive than actual calculation. However, we are limited in a few ways to the size of the chunks:

1. length of individual trajectories in a run
2. memory of a worker process
3. throughput of a worker process

Firstly we are limited in a strong sense by the fact that the largest
possible chunk for a single trajectory is the whole trajectory. Even
if larger chunks would be theoretically, possible. Probably there are
advanced optimizations that could be made if our trajectories happened
to be very numerous and very short, we assume that trajectories are
reasonably long and not numbering in the hundreds of thousands or
millions. This is also dependent on the actual size of a single frame,
which may vary greatly in size for different domains. In any case you
should very likely do some dimensionality reduction before performing
calculations (i.e. stripping out waters in molecular dynamics
simulations).

Secondly, a chunk must be able to fit into the memory of a worker
process. This is a straightforward and intuitive limitation.
Thirdly, we must not make the chunks so large that the degree of
parallelism is diminished.

We leave off here because optimizing parallel calculation is a huge
topic and will change from problem to problem. Thus we encourage trial
and error. In our experience however, a nonresponsive (but non-error
producing calculation) is probably due to the scheduler being
inundated with too many tasks (chunk sizes too small, say of only 1
frame).

"""
import time

from collections import defaultdict
from copy import deepcopy

import numpy as np

import dask.bag as dbag

from wepy.hdf5 import WepyHDF5
from wepy.util.util import concat_traj_fields

RESULT_FIELD_NAME = 'observable'

def traj_fields_chunk_items(wepy_h5_path, fields,
                            run_idxs=Ellipsis,
                            chunk_size=Ellipsis):
    """Generate items that can be used to create a dask.bag object.

    Arguments
    ---------

    wepy_h5_path : str
        The file path to the WepyHDF5 file that will be read from.

    fields : list of str
        The field names/paths for the data to be retrieved.

    chunk_size : int
        This is the size of the chunk (i.e. number of frames) that
        will be retrieved from each trajectory. This is the unit of
        data for which a single task will work on. Dask will also
        partition these chunks as it sees fit.

    Returns
    -------
    chunk_specs : list of dict of str : value

    """

    # open the HDF5
    try:
        wepy_h5 = WepyHDF5(wepy_h5_path, mode='r')
    except OSError:
        print("Failed to open HDF5")
        return None

    with wepy_h5:

        # choose the run idxs
        if run_idxs is not Ellipsis:
            assert all([run_idx in wepy_h5.run_idxs for run_idx in run_idxs]), "run_idx not in runs"
        else:
            run_idxs = wepy_h5.run_idxs

        chunk_specs = []
        for run_idx in run_idxs:
            for traj_idx in wepy_h5.run_traj_idxs(run_idx):

                num_frames = wepy_h5.num_traj_frames(run_idx, traj_idx)

                # determine the specific frame indices in the chunks

                # if the chunk size is either larger than the
                # trajectory, or chunk size is Ellipsis we take the
                # whole trajectory
                if chunk_size is Ellipsis:
                    chunks = [range(num_frames)]
                elif chunk_size > num_frames:
                    chunks = [range(num_frames)]
                else:
                    # split it allowing for an unequal chunk sizes
                    chunks = np.array_split(range(num_frames),
                                            num_frames // chunk_size)

                for frame_idxs in chunks:
                    chunk_spec = {
                        'wepy_h5_path' : wepy_h5_path,
                        'run_idx' : run_idx,
                        'traj_idx' : traj_idx,
                        'frame_idxs' : frame_idxs,
                        'fields' : fields,
                    }
                    chunk_specs.append(chunk_spec)

    return chunk_specs

def load_chunk(chunk_spec):

    wepy_h5 = WepyHDF5(chunk_spec['wepy_h5_path'], mode='r')

    with wepy_h5:
        frame_fields = {}
        for field in chunk_spec['fields']:
            frame_fields[field] = wepy_h5.get_traj_field(chunk_spec['run_idx'],
                                                         chunk_spec['traj_idx'],
                                                         field,
                                                         frames=chunk_spec['frame_idxs'])

    # combine the chunk spec with the traj_fields data
    chunk_spec['traj_fields'] = frame_fields

    return chunk_spec

def chunk_func_funcgen(func,
                       input_keys=['traj_fields'],
                       result_name=None,
                       keep_inputs=False):

    # if the result name wasn't given use the function name
    if result_name is None:
        result_name = func.__name__
    else:
        assert isinstance(result_name, str)

        result_name = result_name


    def chunk_func(chunk_spec):


        assert not result_name in chunk_spec.keys()

        fields = []
        for key in input_keys:

            # pop the trajectory fields off if we are to keep them or not
            if keep_inputs:
                field = chunk_spec[key]
            else:
                field = chunk_spec.pop(key)

            fields.append(field)

        chunk_spec[result_name] = func(*fields)

        return chunk_spec

    return chunk_func

# NOTE: probably shouldn't need this. THere is a dask function called
# 'pluck' which does the same thing, except only for 1 key
def unwrap_chunk_funcgen(keys=None):
    """Consider first using the bag.pluck method"""

    assert keys is not None, "must give at least one key or this is useless"
    assert len(keys) > 0, "must give at least one key or this is useless"

    def unwrap_chunk_func(chunk_struct):

        return {key: value for key, value in chunk_struct.items()
                if key in keys}

    return unwrap_chunk_func

# for reduction
def chunk_key_func(chunk_spec):
    return (chunk_spec['run_idx'], chunk_spec['traj_idx'])

def init_chunk():

    return {'wepy_h5_path' : None,
            'run_idx' : None,
            'traj_idx' : None,
            'fields' : None,
            'frame_idxs' : np.array([])}

def chunk_concat_funcgen(*concat_funcs):

    def chunk_concat(cum_chunk_spec, new_chunk_spec):

        new_chunk = deepcopy(cum_chunk_spec)

        # check to see that the accumulator chunk struct has been
        # initialized, if not initialize it with the values from the new_chunk
        if (
                (cum_chunk_spec['wepy_h5_path'] is None) or
                (cum_chunk_spec['run_idx'] is None) or
                (cum_chunk_spec['traj_idx'] is None) or
                (cum_chunk_spec['fields'] is None)
        ):
            cum_chunk_spec['wepy_h5_path'] = new_chunk_spec['wepy_h5_path']
            cum_chunk_spec['run_idx'] = new_chunk_spec['run_idx']
            cum_chunk_spec['traj_idx'] = new_chunk_spec['traj_idx']
            cum_chunk_spec['fields'] = new_chunk_spec['fields']

        # concatenate the frame indices in this chunk
        new_chunk['frame_idxs'] = np.concatenate(
            [cum_chunk_spec['frame_idxs'], new_chunk_spec['frame_idxs']])

        # for each extra concat function feed it the two chunk specs
        for concat_func in concat_funcs:

            # update the cumulative chunk spec in-place
            new_chunk = concat_func(new_chunk,
                                    new_chunk_spec)


        return new_chunk

    return chunk_concat


def chunk_array_concat_funcgen(field):

    def func(cum_chunk_spec, new_chunk_spec):

        # only add it if it has been initialized in the cum_chunk
        if field in cum_chunk_spec:
            cum_chunk_spec[field] = np.concatenate([cum_chunk_spec[field], new_chunk_spec[field]])

        # otherwise set just the new chunk
        else:
            cum_chunk_spec[field] = new_chunk_spec[field]

        return cum_chunk_spec

    return func

def chunk_traj_fields_concat(cum_chunk_spec, new_chunk_spec):
    """Binary operation for dask foldby reductions for concatenating chunk
    specs with a traj_fields payload"""

    # concatenate the traj fields
    cum_chunk_spec['traj_fields'] = concat_traj_fields(
        [cum_chunk_spec['traj_fields'], new_chunk_spec['traj_fields']])

    return cum_chunk_spec


# this needs to be tested etc.
# def compute_observable_gen(func,
#                            wepy_h5_path,
#                            dask_client,
#                            fields,
#                            chunk_size=Ellipsis,
#                            # TODO replace with traj_sels
#                            run_idxs=Ellipsis):

#     with dask_client:

#         chunks = traj_fields_chunk_items(wepy_h5_path,
#                                          fields,
#                                          chunk_size=chunk_size,
#                                          run_idxs=run_idxs)

#         # DEBUG
#         # TODO add to logging
#         print("generated {} chunks".format(len(chunks)))

#         frame_fields_bag = dbag.from_sequence(chunks)

#         last_step = compute_observable_graph(func, frame_fields_bag, chunk_size)

#         # then we can just iterate through the values (JIT compute)
#         # and set them into the appropriate trajectory
#         while True:
#             traj_id, struct = last_step.take(1)
#             yield (traj_id, struct['observable'])

def _by_traj_to_multidimensional(traj_d):
    """Convert a dictionary of keys (run_idx, traj_idx) and values (of
    dimension val_dim) as arrays to a list of lists of the value
    arrays.

    """

    # get all of the unique run_idxs and sort them
    run_idxs = sorted(list(set([traj_id[0] for traj_id in traj_d.keys()])))

    # then get which trajectories each run has
    run_trajs = defaultdict(list)
    for run_idx, traj_idx in traj_d.keys():
        run_trajs[run_idx].append(traj_idx)

    for run_idx in run_trajs.keys():
        # sort the traj indices (these are unique already within the
        # run)
        run_trajs[run_idx] = sorted(list(run_trajs[run_idx]))

    # then just iterate in order over the run_idxs and inside each the
    # traj indices, as we build these add them to the big structure
    runs_arr = []
    for run_idx in run_idxs:
        run_arr = []
        for traj_idx in run_trajs[run_idx]:
            traj_id = (run_idx, traj_idx)
            run_arr.append(traj_d[traj_id])
        runs_arr.append(run_arr)

    return runs_arr


def compute_observable(func,
                       wepy_h5_path,
                       dask_client,
                       fields,
                       chunk_size=Ellipsis,
                       num_partitions=100,
                       # TODO replace with traj_sels
                       run_idxs=Ellipsis):

    with dask_client:

        chunks = traj_fields_chunk_items(wepy_h5_path,
                                         fields,
                                         chunk_size=chunk_size,
                                         run_idxs=run_idxs)

        # DEBUG
        # TODO add to logging
        print("generated {} chunks".format(len(chunks)))

        frame_fields_bag = dbag.from_sequence(chunks, npartitions=num_partitions)

        last_step = compute_observable_graph(func, frame_fields_bag, chunk_size)

        chunk_results = last_step.compute()

        # get just the result value with the traj id
        results_d = {traj_id : chunk_struct[RESULT_FIELD_NAME]
                   for traj_id, chunk_struct in chunk_results}

        results_arr = _by_traj_to_multidimensional(results_d)

        return results_arr


def compute_observable_graph(func, chunk_bag, chunk_size):

    # since the inputs to any mapped function will be chunk
    # structs (i.e. having metadata about the identity of the
    # chunk, plus any data it drags along, namely traj_fields and
    # computed observables, but perhaps other intermediates in
    # more complex pipelines) we need to wrap any function that we
    # want to call on this data with a function which gets the
    # field of data it needs, and the name of the field it will
    # save data in, optionally we can get rid of the input data if
    # it is no longer needed

    chunk_func = chunk_func_funcgen(func,
                                    input_keys=['traj_fields'],
                                    result_name=RESULT_FIELD_NAME,
                                    keep_inputs=False)

    # start constructing computational graph

    # load chunks into distributed memory
    load_step = chunk_bag.map(load_chunk)

    # run the function over the chunks
    map_step = load_step.map(chunk_func)

    # optionally, defrag the chunks into chunks which are the same
    # as trajectories using a reduce step

    # TODO currently we just always reduce, since that gives us the
    # structure we need at the end, and don't have that for the
    # non-reduction option i.e. when chunk size is same as
    # trajectories. We didn't want to code up the function to
    # structure it that way (in dask) and so leave this for now and if
    # it becomes a performance issue we can make that
    if True: #chunk_size is not Ellipsis:

        # generate the concatenation function for the result so we can
        # reduce and defrag it, this only deals with the part of the chunk of the field name
        result_concat = chunk_array_concat_funcgen(RESULT_FIELD_NAME)

        # now generate the entire chunk concatenation which aggregates
        # multiple other specific field concatenation functions,
        # fundamentally this really only deals with concatenating the
        # frame indices for the chunk
        chunk_concat = chunk_concat_funcgen(result_concat)

        init_cum_chunk = init_chunk()

        # reduce the results by de-fragging the chunks into trajectory
        # chunks
        defrag_step = map_step.foldby(chunk_key_func, chunk_concat,
                                      init_cum_chunk)
    else:
        defrag_step = map_step

    return defrag_step



import logging

from pympler.asizeof import asizeof
import prometheus_client as prom

class SimMonitor:
    """A simulation monitor using a prometheus http server"""

    DEFAULT_PORT = 9001

    def __init__(self,
                 tag="",
                 port=None,
                 reporter_order=(),
    ):

        self.port = port
        self.tag = tag

        self.reporter_order = reporter_order


    def _init_metrics(self):

        logging.info(f"SimMonitor ({self}): Initializing monitoring metrics")

        ## progress tracking
        self.cycle_counter = prom.Counter('wepy_cycle_idx', "",
                                          ['tag']
        )

        # TODO unbinding events

        # TODO largest progresses

        ## object sizes
        self.walker_size_g = prom.Gauge('wepy_walker_single_size_bytes', "",
                                          ['tag']
        )
        self.ensemble_size_g = prom.Gauge('wepy_walker_ensemble_size_bytes', "",
                                          ['tag']
        )

        self.runner_size_g = prom.Gauge('wepy_runner_size_bytes', "",
                                          ['tag']
        )
        self.resampler_size_g = prom.Gauge('wepy_resampler_size_bytes', "",
                                          ['tag']
        )
        self.bc_size_g = prom.Gauge('wepy_bc_size_bytes', "",
                                          ['tag']
        )
        self.mapper_size_g = prom.Gauge('wepy_mapper_size_bytes', "",
                                          ['tag']
        )

        self.sim_manager_size_g = prom.Gauge('wepy_sim_manager_size_bytes', "",
                                          ['tag']
        )

        self.reporter_size_g = prom.Gauge('wepy_reporters_size_bytes',
                                          "",
                                          ['tag', "name"],
        )

        ## timings

        # TODO: convert to Summaries or Histograms for samples

        # components

        self.bc_time_g = prom.Gauge('wepy_bc_cycle_time_seconds', "",
                                          ['tag']
        )
        self.resampling_time_g = prom.Gauge('wepy_resampling_cycle_time_seconds', "",
                                          ['tag']
        )

        # runner splits
        self.runner_precycle_time_g = prom.Gauge('wepy_runner_precycle_time_seconds', "",
                                          ['tag']
        )
        self.runner_postcycle_time_g = prom.Gauge('wepy_runner_postcycle_time_seconds', "",
                                          ['tag']
        )
        self.sim_manager_segment_time_g = prom.Gauge('wepy_sim_manager_segment_time_seconds', "",
                                          ['tag']
        )
        self.sim_manager_segment_overhead_time_g = prom.Gauge(
            'wepy_sim_manager_segment_overhead_time_seconds',
            "",
            ['tag']
        )

        # runner segment splits
        self.runner_segment_gen_sim_time_g = prom.Gauge(
            'wepy_runner_segment_gen_sim_time_seconds',
            "",
            ['tag', 'segment_idx',]
        )

        self.runner_segment_steps_time_g = prom.Gauge(
            'wepy_runner_segment_steps_time_seconds',
            "",
            ['tag', 'segment_idx',]
        )

        self.runner_segment_get_state_time_g = prom.Gauge(
            'wepy_runner_segment_get_state_time_seconds',
            "",
            ['tag', 'segment_idx',]
        )

        self.runner_segment_run_segment_time_g = prom.Gauge(
            'wepy_runner_segment_run_segment_time_seconds',
            "",
            ['tag', 'segment_idx',]
        )

        # work mapper segment times
        self.mapper_seg_times_g = prom.Gauge('wepy_mapper_segment_times_seconds',
                                             "",
                                             ['tag', 'worker_idx', 'seg_idx'])

    def _cleanup_metrics(self):

        logging.info(f"SimMonitor ({self}): cleaning up metrics")

        ## progress tracking
        del self.cycle_counter

        # TODO unbinding events

        # TODO largest progresses

        ## object sizes
        del self.walker_size_g
        del self.ensemble_size_g

        del self.runner_size_g
        del self.resampler_size_g
        del self.bc_size_g
        del self.mapper_size_g

        del self.sim_manager_size_g

        del self.reporter_size_g

        ## timings

        # components

        del self.bc_time_g
        del self.resampling_time_g
        # runner splits
        del self.runner_precycle_time_g
        del self.runner_postcycle_time_g
        del self.sim_manager_segment_time_g
        del self.sim_manager_segment_overhead_time_g
        # runner segment splits
        del self.runner_segment_gen_sim_time_g

        del self.runner_segment_steps_time_g

        del self.runner_segment_get_state_time_g
        del self.runner_segment_run_segment_time_g
        # work mapper segment times
        del self.mapper_seg_times_g

    def init(self,
             port=None,
    ):

        if port is not None:
            port = port

        elif self.port is not None:
            port = self.port

        else:
            port = self.DEFAULT_PORT

        logging.info(f"SimMonitor ({self}): starting prometheus client http server at port {port}")
        prom.start_http_server(port)

        # initialize all the metrics. These need to be done at init
        # time since they can't be pickled
        self._init_metrics()

    def cleanup(self):

        # remove all the metrics
        logging.info(f"SimMonitor ({self}): cleaning up")
        self._cleanup_metrics()



    def cycle_monitor(self, sim_manager, walkers):

        logging.info(f"SimMonitor ({self}): running the cycle monitoring")

        last_report = sim_manager._last_report

        ## Simulation Progress

        # increment the cycle counter
        self.cycle_counter.labels(tag=self.tag).inc()

        ## object sizes

        # get the sizes of the objects
        walker_size = asizeof(walkers[0])
        ensemble_size = asizeof(walkers)

        runner_size = asizeof(sim_manager.runner)
        resampler_size = asizeof(sim_manager.resampler)
        bc_size = asizeof(sim_manager.boundary_conditions)
        mapper_size = asizeof(sim_manager.work_mapper)
        sim_manager_size = asizeof(sim_manager)

        reporter_sizes = {}
        for idx, reporter_name in enumerate(self.reporter_order):
            reporter_sizes[reporter_name] = asizeof(sim_manager.reporters[idx])

        # then update the gauges
        self.walker_size_g.labels(tag=self.tag).set(walker_size)
        self.ensemble_size_g.labels(tag=self.tag).set(ensemble_size)

        self.runner_size_g.labels(tag=self.tag).set(runner_size)
        self.resampler_size_g.labels(tag=self.tag).set(resampler_size)
        self.bc_size_g.labels(tag=self.tag).set(bc_size)
        self.mapper_size_g.labels(tag=self.tag).set(mapper_size)

        self.sim_manager_size_g.labels(tag=self.tag).set(sim_manager_size)

        for reporter_name in self.reporter_order:
            self.reporter_size_g.labels(
                tag=self.tag,
                name=reporter_name,
            ).set(reporter_sizes[reporter_name])


        ## Timings

        self.bc_time_g.labels(tag=self.tag).set(last_report['cycle_bc_time'])
        self.resampling_time_g.labels(tag=self.tag).set(last_report['cycle_resampling_time'])

        # runner splits for each segment
        for seg_idx, segment in enumerate(last_report['runner_splits_time']):

            self.runner_segment_gen_sim_time_g.labels(
                tag=self.tag,
                segment_idx=seg_idx,
            ).set(segment['gen_sim_time'])

            self.runner_segment_steps_time_g.labels(
                tag=self.tag,
                segment_idx=seg_idx,
            ).set(segment['steps_time'])

            self.runner_segment_get_state_time_g.labels(
                tag=self.tag,
                segment_idx=seg_idx,
            ).set(segment['get_state_time'])

            self.runner_segment_run_segment_time_g.labels(
                tag=self.tag,
                segment_idx=seg_idx,
            ).set(segment['run_segment_time'])

        # components
        self.runner_precycle_time_g.labels(
            tag=self.tag
        ).set(
            last_report['runner_precycle_time'])

        self.runner_postcycle_time_g.labels(
            tag=self.tag
        ).set(
            last_report['runner_postcycle_time'])

        self.sim_manager_segment_time_g.labels(
            tag=self.tag
        ).set(
            last_report['cycle_sim_manager_segment_time'])

        self.sim_manager_segment_overhead_time_g.labels(
            tag=self.tag
        ).set(
            last_report['sim_manager_segment_overhead_time'])

        # work mapper workers timings
        for worker_idx, segments in last_report['worker_segment_times'].items():

            for seg_idx, segment in enumerate(segments):
                self.mapper_seg_times_g.labels(
                    tag=self.tag,
                    worker_idx=worker_idx,
                    seg_idx=seg_idx,
                ).set(segment)


        logging.info(f"SimMonitor ({self}): done with cycle monitoring")



import logging

import numpy as np

import mdtraj as mdj

from wepy.reporter.reporter import ProgressiveFileReporter
from wepy.util.mdtraj import json_to_mdtraj_topology, mdtraj_to_json_topology
from wepy.util.json_top import json_top_subset

class WExploreAtomImageReporter(ProgressiveFileReporter):
    """Reporter for generating 3D molecular structures from WExplore
    region images.

    This will only be meaningful for WExplore simulations where the
    region images are actually 3D coordinates.

    """

    FILE_ORDER = ("init_state_path", "image_path")
    SUGGESTED_EXTENSIONS = ("image_top.pdb", "wexplore_images.dcd")


    def __init__(self,
                 init_image=None,
                 image_atom_idxs=None,
                 json_topology=None,
                 **kwargs):
        """Constructor for the WExploreAtomImageReporter.

        Parameters
        ----------

        init_image : numpy.array, optional
            The initial region image. Used for generating the topology
            as well. If not given will be eventually generated.
            (Default = None)

        image_atom_idxs : list of int
            The indices of the atoms that are part of the topology
            subset that comprises the image.

        json_topology : str
            JSON format topology for the whole system. A subset of the
            atoms will be taken using the image_atom_idxs.

        """

        super().__init__(**kwargs)

        assert json_topology is not None, "must give a JSON format topology"
        assert image_atom_idxs is not None, \
            "must give the indices of the atoms for the subset of the topology that is the image"

        self.image_atom_idxs = image_atom_idxs

        self.json_main_rep_top = json_top_subset(json_topology, self.image_atom_idxs)

        self.init_image = None
        self._top_pdb_written = False
        self.image_traj_positions = []

        # if an initial image was given use it, otherwise just don't
        # worry about it, the reason for this is that there is no
        # interface for getting image indices from distance metrics as
        # of now.
        if init_image is not None:
            self.init_image = init_image
            self.image_traj_positions.append(self.init_image)

        # and times
        self.times = [0]


    def init(self, **kwargs):

        super().init(**kwargs)

        if self.init_image is not None:

            image_mdj_topology = json_to_mdtraj_topology(self.json_main_rep_top)

            # initialize the initial image into the image traj
            init_image_traj = mdj.Trajectory([self.init_image],
                                             time=self.times,
                                             topology=image_mdj_topology)



            # save this as a PDB for a topology to view in VMD etc. to go
            # along with the trajectory we will make
            logging.info("Writing initial image to {}".format(self.init_state_path))
            init_image_traj.save_pdb(self.init_state_path)

            self._top_pdb_written = True

    def report(self, cycle_idx=None, resampler_data=None,
               **kwargs):

        # load the json topology as an mdtraj one
        image_mdj_topology = json_to_mdtraj_topology(self.json_main_rep_top)

        # collect the new images defined
        new_images = []
        for resampler_rec in resampler_data:
            image = resampler_rec['image']
            new_images.append(image)

        times = np.array([cycle_idx + 1 for _ in range(len(new_images))])


        # combine the new image positions and times with the old
        self.image_traj_positions.extend(new_images)
        self.times.extend(times)

        # only save if we have an image yet
        if len(self.image_traj_positions) > 0:

            # make a trajectory of the new images, using the cycle_idx as the time
            new_image_traj = mdj.Trajectory(self.image_traj_positions,
                                            time=self.times,
                                            topology=image_mdj_topology)

            # if we haven't already written a topology PDB write it now
            if not self._top_pdb_written:
                new_image_traj[0].save_pdb(self.init_state_path)
                self._top_pdb_written = True

            # then the images to the trajectory file
            new_image_traj.save_dcd(self.image_path)



"""Base class for generating general purpose graphs to external formats.

The primary use for this will be generating GEXF files for use in
visualization.

Routines
--------

LayoutGraph

"""

class LayoutGraph():
    """Base class for networks that will be given layouts for visualization.

    For the graph it wraps it will create a mirror graph which only
    has visualization specific node and edge attributes. This avoids
    namespace clashes between graphs.

    """

    # graphviz colors are given by the RGB or RGBA hex string for them
    GRAPHVIZ_COLOR = 'color'
    GRAPHVIZ_POSITION = 'pos'
    GRAPHVIZ_SHAPE = 'shape'
    GRAPHVIZ_SIZE = 'size'

    GEXF_VIZ = 'viz'
    GEXF_VIZ_COLOR = 'color'
    GEXF_VIZ_COLOR_RED = 'r'
    GEXF_VIZ_COLOR_GREEN = 'g'
    GEXF_VIZ_COLOR_BLUE = 'b'
    GEXF_VIZ_COLOR_ALPHA = 'a'

    GEXF_VIZ_POSITION = 'position'
    GEXF_VIZ_POSITION_X = 'x'
    GEXF_VIZ_POSITION_Y = 'y'
    GEXF_VIZ_POSITION_Z = 'z'

    GEXF_VIZ_SIZE = 'size'
    GEXF_VIZ_SHAPE = 'shape'

    GEXF_EDGE_THICKNESS = 'thickness'
    GEXF_EDGE_SHAPE = 'shape'


    COLOR = GRAPHVIZ_COLOR
    POSITION = GRAPHVIZ_POSITION
    SHAPE = GRAPHVIZ_SHAPE
    SIZE = GRAPHVIZ_SIZE

    def __init__(self, graph):
        """Creates a wrapper for a graph that allows for setting of
        visualization properties and export to external formats without
        modifying the source graph.

        This is done by keeping a mirrored copy of the graph, known as
        the viz_graph, which has all node and edge attributes removed
        but keeps the topology.

        Arguments
        ---------
        graph : any networkx graph


        """

        self._graph = graph

        # we copy the graph to another graph without the attributes so
        # we avoid name collisions between data and visualization metadata
        self._viz_graph = type(self.graph)()
        self._viz_graph.add_nodes_from(self.graph.nodes)
        self._viz_graph.add_edges_from(self.graph.edges)


        # add a "viz" attribute for all nodes for visualization
        # puproses in gexf format
        for node in self.viz_graph.nodes:
            self.viz_graph.node[node][self.GEXF_VIZ] = {}

    @property
    def graph(self):
        """The source data graph the class is wrapping."""
        return self._graph

    @property
    def viz_graph(self):
        """The mirror graph in which visualization attributes are set."""
        return self._viz_graph


    def get_node_attributes(self, attribute_key):
        """Gets attributes from the data graph and returns as a dictionary
        which is compatible with the set_node_viz_attributes methods.

        For instance could be used to get a 'weight' value and set as the color.

        Parameters
        ----------
        attribute_key : str

        Returns
        -------
        node_attributes : dict of node_id: value
            Dictionary mapping of data graph nodes to the value
            specified by the attribute key.

        """

        node_attr_dict = {}
        for node_id in self.graph.nodes:
            node_d = self.graph.node[node_id]
            try:
                value = node_d[attribute_key]

            # if the node doesn't have it set a None instead
            except KeyError:
                value = None

            node_attr_dict[node_id] = value

        return node_attr_dict


    # for setting visualization values of the network in a format agnostic way
    def set_node_viz_attributes(self, attribute_key, node_attribute_dict):
        """Sets attributes of the viz_graph in a format agnostic way.

        Parameters
        ----------
        attribute_key : str
            Key to set node attribute values to.

        node_attribute_dict : dict of node_id: value
            Dictionary mapping node_ids to the value that will be set
            under the attribute key for each node.
        """

        for node_id, value in node_attribute_dict.items():
            self.viz_graph.nodes[node_id][attribute_key] = value

    def set_node_positions(self, node_positions_dict):
        """Set positions of the nodes in the viz graph in a format agnostic way.

        Parameters
        ----------
        node_positions_dict : dict of node_id: position

        """

        self.set_node_viz_attributes(self.POSITION, node_positions_dict)

    def set_node_colors(self, node_colors_dict):
        """Set colors of the nodes in the viz graph in a format agnostic way.

        Parameters
        ----------
        node_colors_dict : dict of node_id: color
        """

        self.set_node_viz_attributes(self.COLOR, node_colors_dict)

    def set_node_sizes(self, node_sizes_dict):
        """Set sizes of the nodes in the viz graph in a format agnostic way.

        Parameters
        ----------
        node_sizes_dict : dict of node_id: int
        """

        self.set_node_viz_attributes(self.SIZE, node_sizes_dict)

    def set_node_shapes(self, node_shapes_dict):
        """Set the shapes of the nodes in the viz graph in a format agnostic way.

        Parameters
        ----------
        node_shapes_dict : dict of node_id: shape_spec
        """

        self.set_node_viz_attributes(self.SHAPE, node_shapes_dict)


    # getters
    def get_node_viz_attributes(self, attribute_key):
        """Get the format agnostic attributes of nodes in the viz graph.

        Parameters
        ----------
        attribute_key : str

        Returns
        -------
        node_attributes : dict of node_id: value
        """

        node_attr_dict = {}
        for node_id in self.viz_graph.nodes:
            node_attr_dict[node_id] = self.graph[node_id][attribute_key]

        return node_attr_dict

    def get_node_positions(self):
        """Get the format agnostic positions of the nodes in the viz graph.

        Returns
        -------
        node_positions : dict of node_id: numpy.ndarray of float
        """


        self.get_node_viz_attributes(self.POSITION)

    def get_node_colors(self):
        """Get the format agnostic colors of the nodes in the viz graph.

        Returns
        -------
        node_colors : dict of node_id: color_spec
        """


        self.get_node_viz_attributes(self.COLOR)

    def get_node_sizes(self):
        """Get the format agnostic size of the nodes in the viz graph.

        Returns
        -------
        node_sizes : dict of node_id: float
        """


        self.get_node_viz_attributes(self.SIZE)

    def get_node_shapes(self):
        """Get the format agnostic shapes of the nodes in the viz graph.

        Returns
        -------
        node_shapes : dict of node_id: shape_spec
        """
        self.get_node_viz_attributes(self.SHAPE)



    @classmethod
    def RGBA_to_hex(cls, color_vec):
        """Convert an RGBA color tuple to the hex representation.

        Parameters
        ----------
        color_vec : array_like of int
            Valid RGB values between 0 and 256

        Returns
        -------
        hex_color : str
            String of the hex value of the color.

        """

        assert not any([True if (color <= 255 and color >=0) else False for color in color_vec]),\
            "invalid color values, must be between 0 and 255"

        return '#' + ''.join(['{:02x}' for _ in color_vec]).format(*[color for color in color_vec])



    # methods for setting gexf visualization attributes
    @classmethod
    def feature_vector_to_gexf_viz_position(cls, coord_vec):
        """Converts an array_like of positions to the corresponding x, y, z
        representation for GEXF representation.

        Parameters
        ----------
        coord_vec : array_like of float of dim (3,)

        Returns
        -------
        gexf_position_spec : dict of str: float
           The node attribute for specifying a position of a node in gexf.

        """

        return {cls.GEXF_VIZ_POSITION_X : float(coord_vec[0]),
                cls.GEXF_VIZ_POSITION_Y : float(coord_vec[1]),
                cls.GEXF_VIZ_POSITION_Z : float(coord_vec[2])}

    @classmethod
    def feature_vector_to_gexf_viz_color_RGB(cls, color_vec):
        """Converts an array_like of positions to the corresponding RGB
        representation for GEXF representation.

        Parameters
        ----------
        color_vec : array_like of int of dim (3,)
            Should be valid RGB values between 0 and 256.

        Returns
        -------
        gexf_color_spec : dict of str: int
           The node attribute for specifying a color of a node in gexf.

        """

        return {cls.GEXF_VIZ_COLOR_RED : int(color_vec[0]),
                cls.GEXF_VIZ_COLOR_GREEN : int(color_vec[1]),
                cls.GEXF_VIZ_COLOR_BLUE : int(color_vec[2])}

    @classmethod
    def feature_vector_to_gexf_viz_color_RGBA(cls, color_vec):
        """Converts an array_like of positions to the corresponding RGBA
        representation for GEXF representation.

        Parameters
        ----------
        color_vec : array_like of int of dim (4,)
            Should be valid RGB values between 0 and 256.

        Returns
        -------
        gexf_color_spec : dict of str: int
           The node attribute for specifying a color of a node in gexf.

        """

        return {cls.GEXF_VIZ_COLOR_RED : int(color_vec[0]),
                cls.GEXF_VIZ_COLOR_GREEN : int(color_vec[1]),
                cls.GEXF_VIZ_COLOR_BLUE : int(color_vec[2]),
                cls.GEXF_VIZ_COLOR_ALPHA : int(color_vec[3])}

    def set_node_gexf_viz(self, viz_key, node_dict):
        """Sets GEXF formatted visualization attributes for all nodes.

        GEXF has a structure where each node has a special compound
        'viz' attribute which can have a number of special attributes
        for shape, position, color, size, etc.

        This method just makes it easy to set one of those attributes
        under all nodes 'viz' attribute.

        Parameters
        ----------
        viz_key : str
            Valid GEXF 'viz' attribute
        node_dict : dict of node_id: value
        """

        for node_id, value in node_dict.items():

            self.viz_graph.node[node_id][self.GEXF_VIZ][viz_key] = value

    def set_node_gexf_positions(self, node_positions_dict):
        """Sets GEXF specific positions attribute for visualization to nodes.

        Parameters
        ----------
        node_positions_dict : dict of node_id: array_like of float of dim (3,)
            Positions for each node.
        """

        # convert the node positions values to a valid dictionary and
        # set the positions viz for all the nodes
        self.set_node_gexf_viz(self.GEXF_VIZ_POSITION,
                          {node_id : self.feature_vector_to_gexf_viz_position(coord)
                           for node_id, coord in node_positions_dict.items()})

    def set_node_gexf_colors_rgb(self, node_colors_dict):
        """Sets GEXF specific RGB color attribute for visualization to nodes.

        Parameters
        ----------
        node_rgb_dict : dict of node_id: array_like of int of dim (3,)
            GEXF RGB color spec for each node.
        """

        # convert the node colors values to a valid dictionary and
        # set the colors viz for all the nodes
        self.set_node_gexf_viz(self.GEXF_VIZ_COLOR,
                               {node_id : self.feature_vector_to_gexf_viz_color_RGB(coord)
                                for node_id, coord in node_colors_dict.items()})

    def set_node_gexf_colors_rgba(self, node_colors_dict):
        """Sets GEXF specific RGBA color attribute for visualization to nodes.

        Parameters
        ----------
        node_rgba_dict : dict of node_id: array_like of int of dim (4,)
            GEXF RGBA color spec for each node.
        """

        # convert the node colors values to a valid dictionary and
        # set the colors viz for all the nodes
        self.set_node_gexf_viz(self.GEXF_VIZ_COLOR,
                               {node_id : self.feature_vector_to_gexf_viz_color_RGBA(coord)
                                for node_id, coord in node_colors_dict.items()})

    def set_node_gexf_sizes(self, node_sizes_dict):
        """Sets GEXF specific node size attribute for visualization to nodes.

        Parameters
        ----------
        node_sizes_dict : dict of node_id: float
            GEXF size spec for each node.
        """

        self.set_node_gexf_viz(self.GEXF_VIZ_SIZE, node_sizes_dict)

    def set_node_gexf_shape(self, node_shape_dict):
        """Sets GEXF specific node shape attribute for visualization to nodes.

        Parameters
        ----------
        node_shapes_dict : dict of node_id: str
            GEXF shape spec for each node.
        """
        self.set_node_gexf_viz(self.GEXF_VIZ_SHAPE, node_shape_dict)


    # TODO: either support or deprecate graphviz. Not sure if this is
    # ready for main since we really only use the gexf one. I was
    # experimenting with this and found it hard.

    # methods for setting graphviz visualization attributes
    # @classmethod
    # def feature_vector_to_graphviz_position(cls, coord_vec):
    #     """

    #     Parameters
    #     ----------
    #     coord_vec :
            

    #     Returns
    #     -------

    #     """

    #     return ','.join(['{}' for _ in coord_vec]).format(*[i for i in coord_vec])


    # @classmethod
    # def feature_vector_to_graphviz_color_RGB(cls, color_vec):
    #     """

    #     Parameters
    #     ----------
    #     color_vec :
            

    #     Returns
    #     -------

    #     """

    #     assert len(color_vec) == 3, "only 3 values for RGB allowed"

    #     return cls.RGBA_to_hex(color_vec)

    # @classmethod
    # def feature_vector_to_graphviz_color_RGBA(cls, color_vec):
    #     """

    #     Parameters
    #     ----------
    #     color_vec :
            

    #     Returns
    #     -------

    #     """

    #     assert len(color_vec) == 4, "only 4 values for RGBA allowed"

    #     return cls.RGBA_to_hex(color_vec)


    # # methods for setting graphviz attributes for visualization

    # def set_node_graphviz(self, viz_key, node_dict):
    #     """

    #     Parameters
    #     ----------
    #     viz_key :
            
    #     node_dict :
            

    #     Returns
    #     -------

    #     """

    #     for node_id, value in node_dict.items():

    #         self.viz_graph.nodes[node_id][viz_key] = value

    # def set_node_graphviz_positions(self, node_positions_dict):
    #     """

    #     Parameters
    #     ----------
    #     node_positions_dict :
            

    #     Returns
    #     -------

    #     """

    #     # convert the node positions values to a valid dictionary and
    #     # set the positions viz for all the nodes
    #     self.set_node_graphviz(self.GRAPHVIZ_POSITION,
    #                       {node_id : self.feature_vector_to_graphviz_position(coord)
    #                        for node_id, coord in node_positions_dict.items()})

    #     # we also pin the nodes here
    #     self.set_node_graphviz('pin',
    #                            {node_id : 'true' for node_id in node_positions_dict.keys()})

    # def set_node_graphviz_colors_rgb(self, node_colors_dict):
    #     """

    #     Parameters
    #     ----------
    #     node_colors_dict :
            

    #     Returns
    #     -------

    #     """

    #     # convert the node colors values to a valid dictionary and
    #     # set the colors viz for all the nodes
    #     self.set_node_graphviz(self.GRAPHVIZ_COLOR,
    #                            {node_id : self.feature_vector_to_graphviz_color_RGB(coord)
    #                                     for node_id, coord in node_colors_dict.items()})

    # def set_node_graphviz_colors_rgba(self, node_colors_dict):
    #     """

    #     Parameters
    #     ----------
    #     node_colors_dict :
            

    #     Returns
    #     -------

    #     """

    #     # convert the node colors values to a valid dictionary and
    #     # set the colors viz for all the nodes
    #     self.set_node_graphviz(self.GRAPHVIZ_COLOR,
    #                            {node_id : self.feature_vector_to_graphviz_color_RGBA_dict(coord)
    #                                     for node_id, coord in node_colors_dict.items()})

    # def set_node_graphviz_sizes(self, node_sizes_dict):
    #     """

    #     Parameters
    #     ----------
    #     node_sizes_dict :
            

    #     Returns
    #     -------

    #     """

    #     # convert the node colors values to a valid dictionary and
    #     # set the colors viz for all the nodes
    #     self.set_node_graphviz(self.GRAPHVIZ_SIZE, node_sizes_dict)

    # def set_node_graphviz_shape(self, node_sizes_dict):
    #     """

    #     Parameters
    #     ----------
    #     node_sizes_dict :
            

    #     Returns
    #     -------

    #     """

    #     # convert the node colors values to a valid dictionary and
    #     # set the colors viz for all the nodes
    #     self.set_node_graphviz(self.GRAPHVIZ_SHAPE, node_sizes_dict)
